from saviialib.libs.log_client.log_client import LogClient
from .types.log_client_types import (
    InfoArgs,
    DebugArgs,
    ErrorArgs,
    WarningArgs,
    LogClientArgs,
    LogStatus,
)

__all__ = [
    "LogClient",
    "InfoArgs",
    "DebugArgs",
    "ErrorArgs",
    "WarningArgs",
    "LogClientArgs",
    "LogStatus",
]



from typing import List, Dict
from saviialib.general_types.error_types.api.saviia_api_error_types import (
    BackupSourcePathError,
)
from saviialib.libs.zero_dependency.utils.booleans_utils import boolean_to_emoji
from saviialib.libs.directory_client import DirectoryClient, DirectoryClientArgs
from saviialib.libs.sharepoint_client import (
    SpListFilesArgs,
)
from saviialib.libs.files_client import (
    WriteArgs,
)

dir_client = DirectoryClient(DirectoryClientArgs(client_name="os_client"))


async def get_pending_files_for_folder(
    sharepoint_client,
    dir_client,
    sharepoint_path: str,
    local_files: set[str],
    failed_files: set[str],
) -> tuple[set[str], str]:
    folders = extract_folders_from_files(local_files)
    sharepoint_files = set()

    async with sharepoint_client:
        for folder in folders:
            files = await sharepoint_client.list_files(
                SpListFilesArgs(f"{sharepoint_path}/{folder}")
            )
            sharepoint_files.update(x["Name"] for x in files["value"])  # type: ignore

    local_basenames = {dir_client.get_basename(f) for f in local_files}
    pending_files = local_basenames.difference(sharepoint_files).union(failed_files)
    summary_msg = (
        f"SharePoint Files: {len(sharepoint_files)}. "
        f"Local Files: {len(local_files)}. "
        f"Failed files: {len(failed_files)}. "
        f"Pending Files: {len(pending_files)}. "
    )
    return pending_files, summary_msg


def parse_execute_response(results: List[Dict]) -> Dict[str, List[str]]:
    try:
        return {
            "new_files": len(
                [item["file_name"] for item in results if item.get("uploaded")]  # type: ignore
            ),
        }
    except (IsADirectoryError, AttributeError, ConnectionError) as error:
        raise BackupSourcePathError(reason=error)


def show_upload_result(uploaded: bool, file_name: str, error_message: str = "") -> str:
    status = boolean_to_emoji(uploaded)
    message = (
        "was uploaded successfully"
        if uploaded
        else f"failed to upload. Error: {error_message}"
    )
    result = f"File {file_name} {message} {status}"
    return result


def calculate_percentage_uploaded(results: List[Dict], total_files: int) -> float:
    uploaded_count = sum(
        1 for result in results if isinstance(result, dict) and result.get("uploaded")
    )
    return (uploaded_count / total_files) * 100 if total_files > 0 else 0


async def count_files_in_directory(base_path: str, folder_name: str) -> int:
    full_path = dir_client.join_paths(base_path, folder_name)
    count = 0
    tree = await dir_client.walk(full_path)
    for root, _, files in tree:
        count += len(files)
    return count


def extract_folders_from_files(files: set[str]) -> set[str]:
    folders = set()
    for f in files:
        parts = f.split("/")
        if len(parts) > 1:
            for i in range(1, len(parts)):
                folders.add("/".join(parts[:i]))
    return folders


async def save_file(files_client, file_name, file_content, mode):
    await files_client.write(
        WriteArgs(
            file_name=file_name,
            file_content=file_content,  # type: ignore
            mode=mode,
        )
    )



import csv
from asyncio import to_thread

from saviialib.libs.directory_client.directory_client import (
    DirectoryClient,
    DirectoryClientArgs,
)
from saviialib.libs.files_client.files_client_contract import FilesClientContract
from saviialib.libs.files_client.types.files_client_types import (
    FilesClientInitArgs,
    ReadArgs,
    WriteArgs,
)


class CsvClient(FilesClientContract):
    def __init__(self, args: FilesClientInitArgs):
        self.dir_client = DirectoryClient(DirectoryClientArgs(client_name="os_client"))

    async def read(self, args: ReadArgs) -> str | bytes:
        raise OSError("This method is not implemented yet.")

    async def write(self, args: WriteArgs) -> None:
        file_type = args.file_name.split(".")[-1]
        file_content = args.file_content
        header = file_content[0].keys()
        if file_type == "tsv":
            delimiter = "\t"
        else:  # Default CSV.
            delimiter = ","

        if args.destination_path == "":
            dest_path = self.dir_client.join_paths(args.file_name)
        else:
            dest_path = self.dir_client.join_paths(
                args.destination_path, args.file_name
            )

        with open(dest_path, "w", newline="") as file:
            writer = csv.DictWriter(file, fieldnames=header, delimiter=delimiter)  # type: ignore
            await to_thread(writer.writeheader)
            await to_thread(writer.writerows, file_content)  # type: ignore



from typing import Any

from saviialib.libs.zero_dependency.utils.datetime_utils import (
    datetime_to_str,
    today,
)


def parse_execute_response(
    thies_fetched_files: dict[str, Any], upload_statistics: dict[str, Any]
) -> dict[str, dict[str, int | str]]:
    return {
        **upload_statistics,
        "processed_files": {
            filename: {
                "file_size": len(data),
                "processed_date": datetime_to_str(today()),
            }
            for filename, data in thies_fetched_files.items()
        },
    }



from abc import ABC, abstractmethod

from .types.ftp_client_types import FtpListFilesArgs, FtpReadFileArgs


class FTPClientContract(ABC):
    @abstractmethod
    async def list_files(self, args: FtpListFilesArgs) -> list[str]:
        pass

    @abstractmethod
    async def read_file(self, args: FtpReadFileArgs) -> bytes:
        pass



from abc import ABC, abstractmethod

from .types.sharepoint_client_types import (
    SpListFilesArgs,
    SpListFoldersArgs,
    SpUploadFileArgs,
    SpCreateFolderArgs,
)


class SharepointClientContract(ABC):
    @abstractmethod
    async def list_files(self, args: SpListFilesArgs) -> list:
        pass

    @abstractmethod
    async def list_folders(self, args: SpListFoldersArgs) -> list:
        pass

    @abstractmethod
    async def upload_file(self, args: SpUploadFileArgs) -> dict:
        pass

    @abstractmethod
    async def create_folder(self, args: SpCreateFolderArgs) -> dict:
        pass



import ftplib
import asyncio
from io import BytesIO
from saviialib.libs.ftp_client.ftp_client_contract import FTPClientContract
from saviialib.libs.ftp_client.types.ftp_client_types import (
    FtpClientInitArgs,
    FtpListFilesArgs,
    FtpReadFileArgs,
)


class FtplibClient(FTPClientContract):
    def __init__(self, args: FtpClientInitArgs) -> None:
        self.host = args.config.ftp_host
        self.port = args.config.ftp_port
        self.password = args.config.ftp_password
        self.user = args.config.ftp_user
        self.client = ftplib.FTP()
        self.args = args

    async def _async_start(self) -> None:
        try:
            await asyncio.to_thread(self.client.connect, self.host, self.port)
            await asyncio.to_thread(self.client.login, self.user, self.password)
        except OSError:
            raise ConnectionRefusedError(
                f"{self.host}:{self.port} isn't active. "
                "Please ensure the server is running and accessible."
            )
        except Exception as error:
            raise ConnectionError(
                f"General connection for {self.host}:{self.port}. {error}"
            )

    async def list_files(self, args: FtpListFilesArgs) -> list[tuple[str, int]]:
        """List files with name and size (like AioFTPClient)."""
        try:
            EXCLUDED_NAMES = [".", ".."]
            await self._async_start()
            await asyncio.to_thread(self.client.cwd, args.path)
            filenames = await asyncio.to_thread(self.client.nlst)
            files_with_sizes = []
            
            for filename in filenames:
                if filename in EXCLUDED_NAMES:
                    continue
                try:
                    size = await asyncio.to_thread(self.client.size, filename)
                except Exception:
                    size = 0
                files_with_sizes.append((filename, int(size))) # type: ignore

            return files_with_sizes
        except Exception as error:
            raise ConnectionAbortedError(error)

    async def read_file(self, args: FtpReadFileArgs) -> bytes:
        await self._async_start()
        try:
            file_content = BytesIO()
            await asyncio.to_thread(
                self.client.retrbinary, "RETR " + args.file_path, file_content.write
            )
            await asyncio.to_thread(file_content.seek, 0)
            file_bytes = await asyncio.to_thread(file_content.read)
            return file_bytes
        except Exception as error:
            raise FileNotFoundError(f"File not found: {args.file_path}") from error



from dataclasses import dataclass, field
from typing import Dict
from logging import Logger


@dataclass
class GetMiniseedFilesUseCaseInput:
    raspberry_shakes: Dict[str, str]
    username: str
    password: str
    ssh_key_path: str
    port: int
    logger: Logger


@dataclass
class GetMiniseedFilesUseCaseOutput:
    download_status: Dict[str, str] = field(default_factory=dict)



from .clients.sharepoint_rest_api import SharepointRestAPI
from .sharepoint_client_contract import SharepointClientContract
from .types.sharepoint_client_types import (
    SharepointClientInitArgs,
    SpListFilesArgs,
    SpListFoldersArgs,
    SpUploadFileArgs,
    SpCreateFolderArgs,
)


class SharepointClient(SharepointClientContract):
    CLIENTS = {"sharepoint_rest_api"}

    def __init__(self, args: SharepointClientInitArgs):
        if args.client_name not in SharepointClient.CLIENTS:
            msg = f"Unsupported client {args.client_name}"
            raise KeyError(msg)
        elif args.client_name == "sharepoint_rest_api":
            self.client_obj = SharepointRestAPI(args)

    @property
    def tenant_id(self):
        return self.client_obj.tenant_id

    @property
    def tenant_name(self):
        return self.client_obj.tenant_name

    @property
    def site_name(self):
        return self.client_obj.site_name

    @property
    def client_id(self):
        return self.client_obj.client_id

    @property
    def client_secret(self):
        return self.client_obj.client_secret

    async def __aenter__(self):
        return await self.client_obj.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.client_obj.__aexit__(exc_type, exc_val, exc_tb)

    async def list_files(self, args: SpListFilesArgs) -> list:
        return await self.client_obj.list_files(args)

    async def list_folders(self, args: SpListFoldersArgs) -> list:
        return await self.client_obj.list_folders(args)

    async def upload_file(self, args: SpUploadFileArgs) -> dict:
        return await self.client_obj.upload_file(args)

    async def create_folder(self, args: SpCreateFolderArgs) -> dict:
        return await self.client_obj.create_folder(args)



from .client.os_client import OsClient
from .directory_client_contract import DirectoryClientContract
from .types.directory_client_types import DirectoryClientArgs
from typing import Iterator


class DirectoryClient(DirectoryClientContract):
    CLIENTS = {"os_client"}

    def __init__(self, args: DirectoryClientArgs) -> None:
        if args.client_name not in DirectoryClient.CLIENTS:
            msg = f"Unsupported client {args.client_name}"
            raise KeyError(msg)

        if args.client_name == "os_client":
            self.client_obj = OsClient()
        self.client_name = args.client_name

    def join_paths(self, *paths: str) -> str:
        return self.client_obj.join_paths(*paths)

    async def path_exists(self, path: str) -> bool:
        return await self.client_obj.path_exists(path)

    async def listdir(self, path: str, more_info: bool = False) -> list:
        return await self.client_obj.listdir(path, more_info)

    async def isdir(self, path: str) -> bool:
        return await self.client_obj.isdir(path)

    async def makedirs(self, path: str) -> None:
        return await self.client_obj.makedirs(path)

    async def remove_file(self, path: str) -> None:
        return await self.client_obj.remove_file(path)

    async def walk(self, path: str) -> Iterator:
        return await self.client_obj.walk(path)

    def relative_path(self, full_path: str, base_folder: str):
        return self.client_obj.relative_path(full_path, base_folder)

    def get_basename(self, path: str):
        return self.client_obj.get_basename(path)



from .clients.aiofiles_client import AioFilesClient
from .clients.csv_client import CsvClient
from .files_client_contract import FilesClientContract
from .types.files_client_types import FilesClientInitArgs, ReadArgs, WriteArgs


class FilesClient(FilesClientContract):
    CLIENTS = {"aiofiles_client", "csv_client"}

    def __init__(self, args: FilesClientInitArgs) -> None:
        if args.client_name not in FilesClient.CLIENTS:
            msg = f"Unsupported client {args.client_name}"
            raise KeyError(msg)

        if args.client_name == "aiofiles_client":
            self.client_obj = AioFilesClient(args)
        elif args.client_name == "csv_client":
            self.client_obj = CsvClient(args)

        self.client_name = args.client_name

    async def read(self, args: ReadArgs):
        return await self.client_obj.read(args)

    async def write(self, args: WriteArgs):
        return await self.client_obj.write(args)



from .sftp_client import SFTPClient
from .types.sftp_client_types import (
    ListfilesArgs,
    DownloadfilesArgs,
    SFTPClientInitArgs,
)

__all__ = ["SFTPClient", "ListfilesArgs", "DownloadfilesArgs", "SFTPClientInitArgs"]



from dataclasses import dataclass
from typing import Any


@dataclass
class SharepointClientInitArgs:
    config: Any
    client_name: str = "sharepoint_rest_api"


@dataclass
class SpListFilesArgs:
    folder_relative_url: str


@dataclass
class SpListFoldersArgs:
    folder_relative_url: str


@dataclass
class SpUploadFileArgs:
    folder_relative_url: str
    file_name: str
    file_content: bytes = bytes()


@dataclass
class SpCreateFolderArgs:
    folder_relative_url: str



from dataclasses import dataclass, field
from typing import Optional, List


@dataclass
class SFTPClientInitArgs:
    client_name: str
    password: Optional[str]
    username: str
    ssh_key_path: str
    host: str = "localhost"
    port: int = 22


@dataclass
class ListfilesArgs:
    path: str


@dataclass
class DownloadfilesArgs:
    source_path: str
    destination_path: str
    files_to_download: List[str] = field(default_factory=list)



from abc import ABC, abstractmethod

from .types.files_client_types import ReadArgs, WriteArgs


class FilesClientContract(ABC):
    @abstractmethod
    async def read(self, args: ReadArgs) -> str | bytes:
        pass

    @abstractmethod
    async def write(self, arg: WriteArgs) -> None:
        pass



from dataclasses import dataclass
from saviialib.general_types.api.saviia_api_types import SharepointConfig
from logging import Logger


@dataclass
class UploadBackupToSharepointUseCaseInput:
    sharepoint_config: SharepointConfig
    local_backup_source_path: str
    sharepoint_destination_path: str
    logger: Logger



from .types.get_miniseed_files_types import (
    GetMiniseedFilesControllerInput,
    GetMiniseedFilesControllerOutput,
)
from saviialib.services.shakes.use_cases.get_miniseed_files import (
    GetMiniseedFilesUseCase,
    GetMiniseedFilesUseCaseInput,
)
from http import HTTPStatus
from saviialib.general_types.error_types.api.saviia_api_error_types import (
    ShakesNoContentError,
)
from saviialib.general_types.error_types.common.common_types import SftpClientError


class GetMiniseedFilesController:
    def __init__(self, input: GetMiniseedFilesControllerInput) -> None:
        self.use_case = GetMiniseedFilesUseCase(
            GetMiniseedFilesUseCaseInput(
                raspberry_shakes=input.raspberry_shakes,
                username=input.config.sftp_user,
                password=input.config.sftp_password,
                ssh_key_path=input.config.ssh_key_path,
                port=input.config.sftp_port,
                logger=input.config.logger,
            )
        )

    async def execute(self) -> GetMiniseedFilesControllerOutput:
        try:
            res = await self.use_case.execute()
            return GetMiniseedFilesControllerOutput(
                message="The MiniSEED files have been downloaded succesfully!",
                status=HTTPStatus.OK.value,
                metadata=res.download_status,
            )
        except ShakesNoContentError as error:
            return GetMiniseedFilesControllerOutput(
                message="No files to upload.",
                status=HTTPStatus.NO_CONTENT.value,
                metadata={"error": error.__str__()},
            )
        except SftpClientError as error:
            return GetMiniseedFilesControllerOutput(
                message="An unexpected error ocurred during SFTP Client connection.",
                status=HTTPStatus.REQUEST_TIMEOUT.value,
                metadata={"error": error.__str__()},
            )



from http import HTTPStatus
from saviialib.general_types.api.saviia_api_types import SharepointConfig

from saviialib.general_types.error_types.api.saviia_api_error_types import (
    BackupUploadError,
    BackupSourcePathError,
    BackupEmptyError,
)
from saviialib.general_types.error_types.common.common_types import (
    EmptyDataError,
    SharepointClientError,
)
from saviialib.services.backup.controllers.types.upload_backup_to_sharepoint_types import (
    UploadBackupToSharepointControllerInput,
    UploadBackupToSharepointControllerOutput,
)
from saviialib.services.backup.use_cases.types.upload_backup_to_sharepoint_types import (
    UploadBackupToSharepointUseCaseInput,
)
from saviialib.services.backup.use_cases.upload_backup_to_sharepoint import (
    UploadBackupToSharepointUsecase,
)


class UploadBackupToSharepointController:
    def __init__(self, input: UploadBackupToSharepointControllerInput):
        self.use_case = UploadBackupToSharepointUsecase(
            UploadBackupToSharepointUseCaseInput(
                sharepoint_config=SharepointConfig(
                    sharepoint_client_id=input.config.sharepoint_client_id,
                    sharepoint_client_secret=input.config.sharepoint_client_secret,
                    sharepoint_site_name=input.config.sharepoint_site_name,
                    sharepoint_tenant_name=input.config.sharepoint_tenant_name,
                    sharepoint_tenant_id=input.config.sharepoint_tenant_id,
                ),
                local_backup_source_path=input.local_backup_source_path,
                sharepoint_destination_path=input.sharepoint_destination_path,
                logger=input.config.logger,
            )
        )

    async def execute(self) -> UploadBackupToSharepointControllerOutput:
        try:
            data = await self.use_case.execute()
            return UploadBackupToSharepointControllerOutput(
                message="Local backup was migrated successfully",
                status=HTTPStatus.OK.value,
                metadata={"data": data},  # type: ignore
            )
        except EmptyDataError:
            return UploadBackupToSharepointControllerOutput(
                message="No files to upload", status=HTTPStatus.NO_CONTENT.value
            )

        except (AttributeError, NameError, ValueError) as error:
            return UploadBackupToSharepointControllerOutput(
                message="An unexpected error occurred during use case initialization.",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )

        except SharepointClientError as error:
            return UploadBackupToSharepointControllerOutput(
                message="Sharepoint Client initialization fails.",
                status=HTTPStatus.INTERNAL_SERVER_ERROR.value,
                metadata={"error": error.__str__()},
            )

        except BackupUploadError as error:
            return UploadBackupToSharepointControllerOutput(
                message="An error occurred during local backup",
                status=HTTPStatus.MULTI_STATUS.value,
                metadata={"error": error.__str__()},
            )

        except BackupSourcePathError as error:
            return UploadBackupToSharepointControllerOutput(
                message="Invalid local backup source path provided.",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )
        except BackupEmptyError as error:
            return UploadBackupToSharepointControllerOutput(
                message="Each folder in the backup folder is empty. Check out again",
                status=HTTPStatus.EXPECTATION_FAILED.value,
                metadata={"error": error.__str__()},
            )



from saviialib.general_types.error_types.api.saviia_api_error_types import (
    SharePointFetchingError,
    SharePointDirectoryError,
    SharePointUploadError,
    ThiesConnectionError,
    ThiesFetchingError,
)
from saviialib.general_types.error_types.common import (
    EmptyDataError,
    FtpClientError,
    SharepointClientError,
)
from saviialib.libs.ftp_client import (
    FTPClient,
    FtpClientInitArgs,
    FtpListFilesArgs,
    FtpReadFileArgs,
)
from saviialib.libs.sharepoint_client import (
    SharepointClient,
    SharepointClientInitArgs,
    SpListFilesArgs,
    SpListFoldersArgs,
    SpUploadFileArgs,
)
from saviialib.libs.directory_client import DirectoryClient, DirectoryClientArgs

from saviialib.libs.files_client import (
    FilesClient,
    FilesClientInitArgs,
    WriteArgs,
    ReadArgs,
)
from saviialib.services.backup.use_cases.types import (
    FtpClientConfig,
    SharepointConfig,
    UpdateThiesDataUseCaseInput,
)
from saviialib.services.backup.utils import (
    parse_execute_response,
)
from saviialib.libs.zero_dependency.utils.datetime_utils import today, datetime_to_str
from .components.create_thies_statistics_file import create_thies_daily_statistics_file
from typing import Set, Dict, List


class UpdateThiesDataUseCase:
    BASE_FOLDER_NAME = "thies"

    def __init__(self, input: UpdateThiesDataUseCaseInput):
        self.sharepoint_client = self._initialize_sharepoint_client(
            input.sharepoint_config
        )
        self.logger = input.logger
        self.thies_ftp_client = self._initialize_thies_ftp_client(input.ftp_config)
        self.sharepoint_folders_path = input.sharepoint_folders_path
        self.ftp_server_folders_path = input.ftp_server_folders_path
        self.local_backup_path = input.local_backup_source_path
        self.sharepoint_base_url = f"/sites/{self.sharepoint_client.site_name}"
        self.uploading = set()
        self.os_client = self._initialize_os_client()
        self.files_client = self._initialize_files_client()

    def _initialize_sharepoint_client(
        self, config: SharepointConfig
    ) -> SharepointClient:
        """Initialize the HTTP client."""
        try:
            return SharepointClient(
                SharepointClientInitArgs(config, client_name="sharepoint_rest_api")
            )
        except ConnectionError as error:
            raise SharepointClientError(error)

    def _initialize_thies_ftp_client(self, config: FtpClientConfig) -> FTPClient:
        """Initialize the FTP client."""
        try:
            return FTPClient(FtpClientInitArgs(config, client_name="ftplib_client"))
        except RuntimeError as error:
            raise FtpClientError(error)

    def _initialize_os_client(self) -> DirectoryClient:
        return DirectoryClient(DirectoryClientArgs(client_name="os_client"))

    def _initialize_files_client(self) -> FilesClient:
        return FilesClient(FilesClientInitArgs(client_name="aiofiles_client"))

    async def _validate_sharepoint_current_folders(self):
        async with self.sharepoint_client:
            folder_base_path = "/".join(
                self.sharepoint_folders_path[0].split("/")[0:-1]
            )
            relative_url = f"{self.sharepoint_base_url}/{folder_base_path}"
            response = await self.sharepoint_client.list_folders(
                SpListFoldersArgs(relative_url)
            )

        current_folders = [item["Name"] for item in response["value"]]  # type: ignore

        for folder_path in self.sharepoint_folders_path:
            folder_name = folder_path.split("/")[-1]
            if folder_name not in current_folders:
                raise SharePointDirectoryError(
                    reason=f"The current folder '{folder_name}' doesn't exist."
                )

    async def fetch_cloud_file_names(self) -> Set[str]:
        """Fetch file names from the RCER cloud."""
        await self._validate_sharepoint_current_folders()
        try:
            cloud_files = set()
            async with self.sharepoint_client:
                for folder_path in self.sharepoint_folders_path:
                    folder_name = folder_path.split("/")[-1]
                    relative_url = f"{self.sharepoint_base_url}/{folder_path}"
                    args = SpListFilesArgs(folder_relative_url=relative_url)
                    response = await self.sharepoint_client.list_files(args)
                    cloud_files.update(
                        {
                            (f"{folder_name}_{item['Name']}", int(item["Length"]))
                            for item in response["value"]  # type: ignore
                        }  # type: ignore
                    )
            return cloud_files
        except Exception as error:
            raise SharePointFetchingError(reason=error)

    async def fetch_thies_file_names(self) -> Set[str]:
        """Fetch file names from the THIES FTP server."""
        try:
            thies_files = set()
            for folder_path in self.ftp_server_folders_path:
                # AV for average, and EXT for extreme.
                prefix = "AVG" if "AV" in folder_path else "EXT"
                files = await self.thies_ftp_client.list_files(
                    FtpListFilesArgs(path=folder_path)
                )
                files_names = {(f"{prefix}_{name}", size) for name, size in files}
                thies_files.update(files_names)
            return thies_files
        except ConnectionRefusedError as error:
            raise ThiesConnectionError(reason=error)
        except ConnectionAbortedError as error:
            raise ThiesFetchingError(reason=error)

    async def fetch_thies_file_content(self) -> Dict[str, bytes]:
        """Fetch the content of files from the THIES FTP server."""
        try:
            content_files = {}
            for file in self.uploading:
                prefix, filename = file.split("_", 1)
                file_path = self.os_client.join_paths(
                    self.local_backup_path,
                    UpdateThiesDataUseCase.BASE_FOLDER_NAME,
                    prefix,
                    filename,
                )
                content = await self.files_client.read(
                    ReadArgs(file_path=file_path, mode="rb")
                )
                self.logger.debug(
                    "[thies_synchronization_lib] Fetching file '%s' from '%s'.",
                    file,
                    file_path,
                )
                # Save file content with its original name.
                content_files[file] = content
            return content_files
        except ConnectionRefusedError as error:
            raise ThiesConnectionError(reason=error)
        except ConnectionAbortedError as error:
            raise ThiesFetchingError(reason=error)

    async def upload_thies_files_to_sharepoint(
        self, files: Dict
    ) -> Dict[str, List[str]]:
        """Upload files to SharePoint and categorize the results."""
        upload_results = {"failed_files": [], "new_files": []}

        async with self.sharepoint_client:
            for file, file_content in files.items():
                try:
                    origin, file_name = file.split("_", 1)
                    # Check if the first folder is for AVG, otherwise assume it's for EXT
                    if "AVG" in self.sharepoint_folders_path[0]:
                        avg_folder = self.sharepoint_folders_path[0]
                        ext_folder = self.sharepoint_folders_path[1]
                    else:
                        avg_folder = self.sharepoint_folders_path[1]
                        ext_folder = self.sharepoint_folders_path[0]
                    folder_path = avg_folder if origin == "AVG" else ext_folder
                    relative_url = f"{self.sharepoint_base_url}/{folder_path}"
                    await self.sharepoint_client.upload_file(
                        SpUploadFileArgs(
                            folder_relative_url=relative_url,
                            file_content=file_content,
                            file_name=file_name,
                        )
                    )
                    upload_results["new_files"].append(file)
                    self.logger.debug(
                        "[thies_synchronization_lib] File '%s' uploaded successfully to '%s' ✅",
                        file_name,
                        relative_url,
                    )

                except ConnectionError as error:
                    self.logger.error(
                        "[thies_synchronization_lib] Unexpected error from with file '%s'",
                        file_name,
                    )
                    upload_results["failed_files"].append(
                        f"{file} (Error: {str(error)})"
                    )

        if upload_results["failed_files"]:
            raise SharePointUploadError(
                reason="Files failed to upload: "
                + ", ".join(upload_results["failed_files"])
            )

        return upload_results

    async def _sync_pending_files(self, thies_files: set, cloud_files: set) -> set:
        thies_files_dict = {name: size for name, size in thies_files}
        cloud_files_dict = {name: size for name, size in cloud_files}
        uploading = set()
        for f_from_thies, f_size_from_thies in thies_files_dict.items():
            # If is in thies but not in cloud, then upload it
            if f_from_thies not in cloud_files_dict:
                uploading.add(f_from_thies)
            else:
                # If the file is in both services, but the size is different, then upload it
                f_size_from_cloud = cloud_files_dict[f_from_thies]
                if f_size_from_thies != f_size_from_cloud:
                    uploading.add(f_from_thies)
        return uploading

    async def _extract_thies_daily_statistics(self) -> None:
        # Read the daily files and save each data in the folder
        daily_files = [
            prefix + datetime_to_str(today(), date_format="%Y%m%d") + ".BIN"
            for prefix in ["AVG_", "EXT_"]
        ]
        # Receive from FTP server and  write the file in thies-daily-files
        for file in daily_files:
            prefix, filename = file.split("_", 1)
            # The first path is for AVG files. The second file is for EXT files
            file_path = self.os_client.join_paths(
                self.local_backup_path, UpdateThiesDataUseCase.BASE_FOLDER_NAME, prefix
            )
            files = await self.os_client.listdir(file_path)
            if filename not in files:
                reason = "The file might not be available yet for statistics."
                self.logger.warning("[thies_synchronization_lib] Warning: %s", reason)
                self.logger.warning(
                    "[thies_synchronization_lib] Skipping the creation of daily statistics %s",
                    filename,
                )
                return
        await create_thies_daily_statistics_file(
            self.local_backup_path, self.os_client, self.logger
        )

    async def _validate_local_backup(self):
        backup_path = self.os_client.join_paths(
            self.local_backup_path, UpdateThiesDataUseCase.BASE_FOLDER_NAME
        )
        backup_dir_exists = await self.os_client.path_exists(backup_path)
        if not backup_dir_exists:
            await self.os_client.makedirs(backup_path)

        for dest_folder in {"AVG", "EXT"}:
            dest_folder_path = self.os_client.join_paths(backup_path, dest_folder)
            dest_folder_exists = await self.os_client.path_exists(dest_folder_path)
            if not dest_folder_exists:
                await self.os_client.makedirs(
                    self.os_client.join_paths(backup_path, dest_folder)
                )

    async def _fill_local_backup(self, thies_files: Set[str]) -> None:
        local_avg_files = await self.os_client.listdir(
            self.os_client.join_paths(
                self.local_backup_path, UpdateThiesDataUseCase.BASE_FOLDER_NAME, "AVG"
            ),
            more_info=True,
        )
        local_avg_files = {filename: size for filename, size in local_avg_files}
        local_ext_files = await self.os_client.listdir(
            self.os_client.join_paths(
                self.local_backup_path, UpdateThiesDataUseCase.BASE_FOLDER_NAME, "EXT"
            ),
            more_info=True,
        )
        local_ext_files = {filename: size for filename, size in local_ext_files}
        try:
            for file, orig_size in thies_files:
                prefix, filename = file.split("_", 1)
                # The first path is for AVG files. The second file is for EXT files
                folder_path = next(
                    (
                        path
                        for path in self.ftp_server_folders_path
                        if prefix == ("AVG" if "AV" in path else "EXT")
                    ),
                    self.ftp_server_folders_path[0],  # Default to the first path
                )
                dest_path = self.os_client.join_paths(
                    self.local_backup_path,
                    UpdateThiesDataUseCase.BASE_FOLDER_NAME,
                    prefix,
                )
                new_size = (
                    local_avg_files.get(filename, None)
                    if prefix == "AVG"
                    else local_ext_files.get(filename, None)
                )
                should_be_added = False
                if new_size and new_size != orig_size:
                    should_be_added = True
                elif not new_size:
                    should_be_added = True
                else:
                    # Should not be added, because it has the same size and it's saved in the local dir.
                    pass

                if not should_be_added:
                    continue

                self.logger.debug(
                    f"[thies_synchronization_lib] Saving {filename} in Thies local backup"
                )
                file_path = f"{folder_path}/{filename}"
                file_content = await self.thies_ftp_client.read_file(
                    FtpReadFileArgs(file_path)
                )
                await self.files_client.write(
                    WriteArgs(
                        file_name=filename,
                        file_content=file_content,
                        mode="wb",
                        destination_path=dest_path,
                    )
                )
        except ConnectionRefusedError as error:
            raise ThiesConnectionError(reason=error)
        except ConnectionAbortedError as error:
            raise ThiesFetchingError(reason=error)

    async def execute(self):
        """Synchronize data from the THIES Center to the cloud."""
        self.logger.debug("[thies_synchronization_lib] Starting ...")
        await self._validate_local_backup()
        try:
            thies_files = await self.fetch_thies_file_names()
            await self._fill_local_backup(thies_files)
        except RuntimeError as error:
            raise FtpClientError(error)
        self.logger.debug(
            "[thies_synchronization_lib] Total files fetched from THIES: %s",
            str(len(thies_files)),
        )
        try:
            cloud_files = await self.fetch_cloud_file_names()
        except RuntimeError as error:
            raise SharepointClientError(error)  # type: ignore
        self.logger.debug(
            "[thies_synchronization_lib] Total files fetched from Sharepoint: %s",
            str(len(cloud_files)),
        )
        self.uploading = await self._sync_pending_files(thies_files, cloud_files)
        # Extract thies statistics for SAVIIA Sensors
        await self._extract_thies_daily_statistics()
        if not self.uploading:
            raise EmptyDataError(reason="No files to upload.")
        self.logger.debug(
            "[thies_synchronization_lib] Total new files to upload: %s",
            str(len(self.uploading)),
        )
        # Fetch the content of the files to be uploaded from THIES FTP Server
        thies_fetched_files = await self.fetch_thies_file_content()
        # # Upload the fetched files to SharePoint and gather statistics
        upload_statistics = await self.upload_thies_files_to_sharepoint(
            thies_fetched_files
        )
        self.logger.info(upload_statistics)
        self.logger.debug(
            "[thies_synchronization_lib] All the files were uploaded successfully 🎉"
        )

        return parse_execute_response(thies_fetched_files, upload_statistics)  # type: ignore



from datetime import datetime
from zoneinfo import ZoneInfo


def today(timezone: str = "America/Santiago") -> datetime:
    """
    Return the current date.

    :param timezone: A string representing the IANA timezone name.
        Defaults to "America/Santiago".
    :return datetime:
    """
    return datetime.now(tz=ZoneInfo(timezone))


def datetime_to_str(date: datetime, date_format: str = "%m/%d/%Y, %H:%M:%S") -> str:
    """
    Convert a datetime object to a string in the specified format.

    :param date: The datetime object to convert.
    :param date_format: The format to convert the datetime object to.
        Defaults to "%Y-%m-%dT%H:%M:%S".
    :return: A string in the specified format.
    """
    return date.strftime(date_format)



from dataclasses import dataclass
from logging import Logger


@dataclass
class SaviiaBackupConfig:
    """
    Configuration for backing up files to SharePoint.

    Attributes:
        sharepoint_client_id (str): Client ID for SharePoint authentication.
        sharepoint_client_secret (str): Client secret for SharePoint authentication.
        sharepoint_tenant_id (str): Tenant ID for SharePoint authentication.
        sharepoint_tenant_name (str): Tenant name for SharePoint.
        sharepoint_site_name (str): Site name in SharePoint.
        local_backup_source_path (str): Local path to backup.
    """

    sharepoint_client_id: str
    sharepoint_client_secret: str
    sharepoint_tenant_id: str
    sharepoint_tenant_name: str
    sharepoint_site_name: str
    logger: Logger



from dataclasses import dataclass, field
from typing import Dict, List
from saviialib.general_types.api.saviia_api_types import (
    FtpClientConfig,
    SharepointConfig,
)
from logging import Logger


@dataclass
class UpdateThiesDataUseCaseInput:
    ftp_config: FtpClientConfig
    sharepoint_config: SharepointConfig
    sharepoint_folders_path: List
    ftp_server_folders_path: List
    local_backup_source_path: str
    logger: Logger


@dataclass
class UpdateThiesDataUseCaseOutput:
    message: str
    status: int = 0
    metadata: Dict[str, str] = field(default_factory=dict)



from .sftp_client_contract import SFTPClientContract
from .types.sftp_client_types import (
    SFTPClientInitArgs,
    ListfilesArgs,
    DownloadfilesArgs,
)
from .clients.asyncssh_sftp_client import AsyncsshSFTPClient
from typing import List


class SFTPClient(SFTPClientContract):
    CLIENTS = {"asyncssh_sftp"}

    def __init__(self, args: SFTPClientInitArgs) -> None:
        if args.client_name not in SFTPClient.CLIENTS:
            msg = f"Unsupported client {args.client_name}"
            raise KeyError(msg)
        if args.client_name == "asyncssh_sftp":
            self.client_obj = AsyncsshSFTPClient(args)
        self.client_name = args.client_name

    async def list_files(self, args: ListfilesArgs) -> List[str]:
        return await self.client_obj.list_files(args)

    async def download_files(self, args: DownloadfilesArgs) -> None:
        return await self.client_obj.download_files(args)



from .controllers.get_miniseed_files import (
    GetMiniseedFilesController,
    GetMiniseedFilesControllerInput,
)
from saviialib.general_types.api.saviia_shakes_api_types import SaviiaShakesConfig

from typing import Dict


class ShakesAPI:
    """This class provides methods for interacting with Raspberry Shakes"""

    def __init__(self, config: SaviiaShakesConfig) -> None:
        self.config = config

    async def get_miniseed_files(self, raspberry_shakes: Dict[str, str]):
        """Download the MiniSEED files from the SFTP Server provided by each Raspberry Shake.
        Args:
            raspberry_shakes (dict): Dictionary where the key is the name of the Raspberry Shake,
                and the value is the IP Address.
        Returns:
            response (dict): A dictionary containg the response from the download operation.
                This response will tipically include the message, the response status, and metadata.
        """
        controller = GetMiniseedFilesController(
            GetMiniseedFilesControllerInput(
                config=self.config, raspberry_shakes=raspberry_shakes
            )
        )
        response = await controller.execute()
        return response.__dict__



from .clients.aioftp_client import AioFTPClient
from .clients.ftplib_client import FtplibClient
from .ftp_client_contract import FTPClientContract
from .types.ftp_client_types import FtpClientInitArgs, FtpListFilesArgs, FtpReadFileArgs


class FTPClient(FTPClientContract):
    CLIENTS = {"aioftp_client", "ftplib_client"}

    def __init__(self, args: FtpClientInitArgs) -> None:
        if args.client_name not in FTPClient.CLIENTS:
            msg = f"Unsupported client {args.client_name}"
            raise KeyError(msg)

        if args.client_name == "aioftp_client":
            self.client_obj = AioFTPClient(args)
        elif args.client_name == "ftplib_client":
            self.client_obj = FtplibClient(args)
        self.client_name = args.client_name

    async def list_files(self, args: FtpListFilesArgs) -> list[str]:
        return await self.client_obj.list_files(args)

    async def read_file(self, args: FtpReadFileArgs) -> bytes:
        return await self.client_obj.read_file(args)



import asyncio
from time import time
from logging import Logger
from saviialib.general_types.error_types.api.saviia_api_error_types import (
    BackupEmptyError,
    BackupSourcePathError,
    BackupUploadError,
)
from saviialib.general_types.error_types.common import (
    SharepointClientError,
)
from saviialib.libs.directory_client import DirectoryClient, DirectoryClientArgs
from saviialib.libs.files_client import (
    FilesClient,
    FilesClientInitArgs,
    ReadArgs,
)
from saviialib.libs.sharepoint_client import (
    SharepointClient,
    SharepointClientInitArgs,
    SpUploadFileArgs,
    SpCreateFolderArgs,
)
from saviialib.libs.log_client import (
    LogClient,
    LogClientArgs,
    InfoArgs,
    DebugArgs,
    WarningArgs,
    ErrorArgs,
    LogStatus,
)

from saviialib.services.backup.utils.upload_backup_to_sharepoint_utils import (
    calculate_percentage_uploaded,
    count_files_in_directory,
    parse_execute_response,
    show_upload_result,
    get_pending_files_for_folder,
    save_file,
)
from saviialib.libs.zero_dependency.utils.booleans_utils import boolean_to_emoji

from .types.upload_backup_to_sharepoint_types import (
    UploadBackupToSharepointUseCaseInput,
)
from typing import Dict, List


class UploadBackupToSharepointUsecase:
    LOCAL_BACKUP_NAME = "saviia-local-backup"

    def __init__(self, input: UploadBackupToSharepointUseCaseInput):
        self.sharepoint_config = input.sharepoint_config
        self.local_backup_source_path = input.local_backup_source_path
        self.sharepoint_destination_path = input.sharepoint_destination_path
        self.files_client = self._initialize_files_client()
        self.dir_client = self._initialize_directory_client()
        self.sharepoint_client = self._initalize_sharepoint_client()

        self.grouped_files_by_folder = {}
        self.logger: Logger = input.logger
        self.infofile: Dict[str, str | List[str]] = {}
        self.log_client = LogClient(
            LogClientArgs(
                client_name="logging",
                class_name="backup_to_sharepoint",
                service_name="backup",
                active_record=True,
            )
        )

    def _initalize_sharepoint_client(self):
        return SharepointClient(
            SharepointClientInitArgs(
                self.sharepoint_config, client_name="sharepoint_rest_api"
            )
        )

    def _initialize_directory_client(self):
        return DirectoryClient(DirectoryClientArgs(client_name="os_client"))

    def _initialize_files_client(self):
        return FilesClient(FilesClientInitArgs(client_name="aiofiles_client"))

    async def _group_files_by_folder(self) -> dict[str, list[str]]:
        """Groups files to upload by their parent folder"""
        folder_names = await self.dir_client.listdir(self.local_backup_source_path)
        grouped = {}
        if len(folder_names) == 0:
            raise BackupEmptyError

        async def walk_directory(base_folder: str) -> List[str]:
            """Recursivelty collect file paths relative to base folder"""
            all_files = []
            entries = await self.dir_client.listdir(base_folder)
            if ".PASS.txt" in entries:
                return []
            for entry in entries:
                full_path = self.dir_client.join_paths(base_folder, entry)
                if await self.dir_client.isdir(full_path):
                    sub_entries = await self.dir_client.listdir(full_path)
                    if ".PASS.txt" in sub_entries:
                        continue
                    sub_files = await walk_directory(full_path)
                    all_files.extend(sub_files)
                else:
                    rel_path = self.dir_client.relative_path(
                        full_path, self.local_backup_source_path
                    )
                    all_files.append(rel_path)
            return all_files

        for folder_name in folder_names:
            folder_path = self.dir_client.join_paths(
                self.local_backup_source_path, folder_name
            )
            if not await self.dir_client.isdir(folder_path):
                continue
            entries = await self.dir_client.listdir(folder_path)
            if ".PASS.txt" in entries:
                grouped[folder_name] = set()
                continue

            all_files = await walk_directory(folder_path)
            grouped[folder_name] = set(all_files)
        return grouped

    async def _create_or_update_infofile(self):
        """Creates or updates .infofile.json with pass/reset flags for each directory."""
        self.log_client.method_name = "_create_or_update_infofile"
        self.log_client.info(
            InfoArgs(LogStatus.STARTED, metadata={"msg": "Creating .infofile.json"})
        )
        infofile_path = self.dir_client.join_paths(
            self.local_backup_source_path, ".infofile.json"
        )
        infofile_exists = await self.dir_client.path_exists(infofile_path)

        if infofile_exists:
            self.log_client.debug(
                DebugArgs(LogStatus.ALERT, metadata={"msg": "Updating .infofile.json"})
            )
            self.infofile = await self.files_client.read(
                ReadArgs(file_path=infofile_path, mode="json")
            )  # type: ignore
            return

        for folder, files in self.grouped_files_by_folder.items():
            should_pass = len(files) == 0
            should_reset = any(
                f.endswith(".RESET.txt") or "/.RESET.txt" in f for f in files
            )
            self.infofile[folder] = {  # type: ignore
                "pass": should_pass,
                "reset": should_reset,
                "failed": [],
            }
        self.log_client.info(
            InfoArgs(
                LogStatus.SUCCESSFUL, metadata={"msg": "Infofile created succesfully"}
            )
        )
        for dir in self.infofile:
            self.log_client.debug(
                DebugArgs(
                    LogStatus.ALERT,
                    metadata={
                        "msg": (
                            f"{dir}: Pass {boolean_to_emoji(self.infofile[dir]['pass'])} "  # type: ignore
                            f"Reset {boolean_to_emoji(self.infofile[dir]['reset'])}"  # type: ignore
                        )
                    },
                )
            )

    async def prepare_backup(self):
        self.log_client.method_name = "prepare_backup"
        self.log_client.debug(
            DebugArgs(
                LogStatus.STARTED,
                metadata={"msg": "* Extracting folders from local backup directory *"},
            )
        )
        # Check if the local backup directory exists. If doesn't exist, then create it.
        local_backup_path_exists = await self.dir_client.path_exists(
            self.local_backup_source_path
        )
        if not local_backup_path_exists:
            raise BackupSourcePathError(
                reason=f"'{self.local_backup_source_path}' doesn't exist."
            )
        # Create the destination directory if it doesn't exist
        complete_destination_path = (
            self.sharepoint_destination_path
            + "/"
            + UploadBackupToSharepointUsecase.LOCAL_BACKUP_NAME
        )
        async with self.sharepoint_client:
            await self.sharepoint_client.create_folder(
                SpCreateFolderArgs(folder_relative_url=complete_destination_path)
            )
        # Check out the directories and the files inside each one
        self.grouped_files_by_folder = await self._group_files_by_folder()
        # Create or read the info file
        await self._create_or_update_infofile()
        # Replicate local directories in the sharepoint directory.
        for folder in self.grouped_files_by_folder.keys():
            should_pass = self.infofile[folder]["pass"]  # type: ignore
            should_reset = self.infofile[folder]["reset"]  # type: ignore
            if should_reset:
                self.log_client.warning(
                    WarningArgs(
                        LogStatus.ALERT,
                        metadata={
                            "msg": f"The '{folder}' will reseted before the synchronization as it contains the .RESET.txt file ❗️"
                        },
                    )
                )
            if should_pass:
                self.log_client.debug(
                    DebugArgs(
                        LogStatus.ALERT,
                        metadata={
                            "msg": f"The '{folder}' will not be synchronised as it contains the .PASS.txt file 👻"
                        },
                    )
                )
                continue
            async with self.sharepoint_client:
                await self.sharepoint_client.create_folder(
                    SpCreateFolderArgs(
                        folder_relative_url=complete_destination_path + "/" + folder
                    )
                )
            self.grouped_files_by_folder[folder] = {  # type: ignore
                f
                for f in self.grouped_files_by_folder[folder]
                if not f.endswith(".PASS.txt") and not f.endswith(".RESET.txt")
            }
        self.sharepoint_destination_path = complete_destination_path

        self.log_client.debug(
            DebugArgs(LogStatus.SUCCESSFUL, metadata={"msg": "Ready to migrate"})
        )

    async def migrate_files(self) -> list:
        self.log_client.method_name = "migrate_files"
        self.log_client.debug(
            DebugArgs(
                LogStatus.STARTED, metadata={"msg": "* Counting files to migrate *"}
            )
        )
        tasks = []
        local_files: set
        for folder_name, local_files in self.grouped_files_by_folder.items():  # type: ignore
            should_pass = self.infofile[folder_name]["pass"]  # type: ignore
            if should_pass:
                continue

            count_files_in_dir = await count_files_in_directory(
                self.local_backup_source_path, folder_name
            )
            if count_files_in_dir == 0:
                self.log_client.debug(
                    DebugArgs(
                        LogStatus.ALERT,
                        metadata={"msg": f"The folder '{folder_name}' is empty ⚠️"},
                    )
                )
                continue
            # Check out the pending files and add the failed failes
            self.log_client.debug(
                DebugArgs(
                    LogStatus.ALERT,
                    metadata={
                        "msg": f"Checking files to upload in '{folder_name}' directory 👀"
                    },
                )
            )
            async with self.sharepoint_client:
                pending_files, summary_msg = await get_pending_files_for_folder(
                    self.sharepoint_client,
                    self.dir_client,
                    self.sharepoint_destination_path,
                    local_files,
                    set(self.infofile[folder_name]["failed"]),  # type: ignore
                )
                self.log_client.debug(
                    DebugArgs(
                        LogStatus.ALERT,
                        metadata={"msg": summary_msg},
                    )
                )
                # Update the files to upload
                self.grouped_files_by_folder[folder_name] = [
                    f
                    for f in local_files
                    if self.dir_client.get_basename(f) in pending_files
                ]

                if len(pending_files) == 0:
                    self.log_client.debug(
                        DebugArgs(
                            LogStatus.ALERT,
                            metadata={
                                "msg": f"All the files in '{folder_name}' have already been synchronised"
                            },
                        )
                    )
                    continue
            # Export the file to Sharepoint
            for file_name in self.grouped_files_by_folder[folder_name]:  # type: ignore
                if file_name in [".RESET.txt", ".PASS.txt"]:
                    continue
                tasks.append(self._upload_file_to_sharepoint(folder_name, file_name))
        self.log_client.debug(
            DebugArgs(
                LogStatus.SUCCESSFUL,
                metadata={"msg": "* The reviewing process has finished *"},
            )
        )
        return tasks

    async def _upload_file_to_sharepoint(self, folder_name, file_name) -> dict:
        """Task for uploads a file and logs progress."""
        self.log_client.method_name = "_upload_file_to_sharepoint"
        self.log_client.debug(
            DebugArgs(
                LogStatus.STARTED,
                metadata={"msg": f"Uploading file '{file_name}' from '{folder_name}' "},
            )
        )
        # Retrieve the content from the file in the local directory
        file_path = self.dir_client.join_paths(self.local_backup_source_path, file_name)
        file_content = await self.files_client.read(ReadArgs(file_path, mode="rb"))
        # Upload the local file to Sharepoint directory
        uploaded, error_message = None, ""
        try:
            sharepoint_client = SharepointClient(
                SharepointClientInitArgs(
                    self.sharepoint_config, client_name="sharepoint_rest_api"
                )
            )
        except ConnectionError as error:
            self.log_client.error(
                ErrorArgs(
                    LogStatus.ERROR,
                    metadata={"msg": error.__str__()},
                )
            )
            raise SharepointClientError(error)
        async with sharepoint_client:
            try:
                relative_folder = "/".join(file_name.split("/")[:-1])
                folder_url = (
                    f"{self.sharepoint_destination_path}/{relative_folder}"
                    if relative_folder
                    else self.sharepoint_destination_path
                )
                await sharepoint_client.create_folder(
                    SpCreateFolderArgs(folder_relative_url=folder_url)
                )

                await sharepoint_client.upload_file(
                    SpUploadFileArgs(
                        folder_relative_url=folder_url,
                        file_content=file_content,  # type: ignore
                        file_name=self.dir_client.get_basename(file_name),
                    )
                )
                uploaded = True
                # Remove the file from the source directory if RESET is true
                should_reset = self.infofile[folder_name]["reset"]  # type: ignore
                if should_reset:
                    source_file_path = self.dir_client.join_paths(
                        self.local_backup_source_path, file_name
                    )
                    await self.dir_client.remove_file(source_file_path)
                    self.log_client.debug(
                        DebugArgs(
                            LogStatus.ALERT,
                            metadata={
                                "msg": f"File {file_name} has been deleted from '{folder_name}'"
                            },
                        )
                    )
            except ConnectionError as error:
                error_message = str(error)
                uploaded = False
                # Add the file to failed failes
                self.log_client.debug(
                    DebugArgs(
                        LogStatus.ALERT,
                        metadata={"msg": f"{file_name} could not be uploaded 🚨"},
                    )
                )
                self.infofile[folder_name]["failed"].append(file_name)  # type: ignore
        self.log_client.debug(
            DebugArgs(
                LogStatus.SUCCESSFUL if uploaded else LogStatus.ALERT,
                metadata={
                    "msg": show_upload_result(uploaded, file_name, error_message)
                },
            )
        )
        return {
            "parent_folder": folder_name,
            "file_name": file_name,
            "uploaded": uploaded,
            "error_message": error_message,
        }

    async def execute(self):
        """Exports all files from the local backup folder to SharePoint cloud."""
        self.log_client.method_name = "execute"
        self.log_client.info(
            InfoArgs(LogStatus.STARTED, metadata={"msg": "Starting backup process"})
        )
        start_time = time()
        await self.prepare_backup()
        tasks = await self.migrate_files()
        already_migrated = (
            len(tasks) == 0
        )  # If the tasks set is zero, then were no any errors.
        results = await asyncio.gather(*tasks, return_exceptions=True)
        success = calculate_percentage_uploaded(results, len(tasks))  # type: ignore
        if success < 100.0 and not already_migrated:
            self.log_client.method_name = "execute"
            self.log_client.warning(
                WarningArgs(
                    LogStatus.FAILED,
                    metadata={
                        "msg": (
                            "Not all the files have been uploaded ⚠️."
                            f"Files failed to upload: {(1 - success):.2%}"
                        )
                    },
                )
            )
            # Save any failed files for exporting in the next migration
            await save_file(self.files_client, ".infofile.json", self.infofile, "json")
            await save_file(
                self.files_client,
                "BACKUP_LOG_HISTORY.log",
                "\n".join(self.log_client.log_history),
                "w",
            )
            raise BackupUploadError(
                reason="Not all the files have been uploaded successfully."
            )
        else:
            end_time = time()
            backup_time = end_time - start_time
            self.log_client.method_name = "execute"
            self.log_client.debug(
                DebugArgs(
                    LogStatus.SUCCESSFUL,
                    metadata={
                        "msg": (
                            f"Migration time: {backup_time:.2f} seconds ✨. "
                            "All the files were uploaded successfully 🎉"
                        )
                    },
                )
            )
            await save_file(self.files_client, ".infofile.json", self.infofile, "json")
            await save_file(
                self.files_client,
                "BACKUP_LOG_HISTORY.log",
                "\n".join(self.log_client.log_history),
                "w",
            )
            return parse_execute_response(results)  # type: ignore



class EmptyDataError(Exception):
    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return "The data provided is empty. " + self.reason.__str__()


class SharepointClientError(Exception):
    def __str__(self):
        return "SharePoint API REST Client initialization fails."


class FtpClientError(Exception):
    def __str__(self):
        return "Ftp Client initialization fails."


class SftpClientError(Exception):
    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return "SFTP Client initialization fails." + self.reason.__str__()



from dataclasses import dataclass
from typing import Literal, Union, List, Dict


@dataclass
class FilesClientInitArgs:
    client_name: str = "aiofiles_client"


@dataclass
class ReadArgs:
    """
    Represents the arguments required to read a file.

    Attributes:
        file_path (str): The path to the file to be read.
        mode (Literal['r', 'rb']): The mode in which the file should be opened.
            - 'r': Open for reading (text mode).
            - 'rb': Open for reading (binary mode).
    """

    file_path: str
    mode: Literal["r", "rb", "json"]
    encoding: str = "utf-8"


@dataclass
class WriteArgs:
    file_name: str
    file_content: Union[str, bytes, List[Dict]]
    mode: Literal["w", "wb", "a", "json"]
    destination_path: str = ""



from .types.get_media_files_types import (
    GetMediaFilesControllerInput,
    GetMediaFilesControllerOutput,
)
from saviialib.services.netcamera.use_cases.get_media_files import (
    GetMediaFilesUseCase,
    GetMediaFilesUseCaseInput,
)
from http import HTTPStatus
from saviialib.general_types.error_types.api.saviia_netcamera_error_types import (
    NetcameraConnectionError,
)


class GetMediaFilesController:
    def __init__(self, input: GetMediaFilesControllerInput) -> None:
        self.use_case = GetMediaFilesUseCase(
            GetMediaFilesUseCaseInput(
                cameras=input.cameras,
                username=input.config.username,
                password=input.config.password,
                protocol=input.config.protocol,
                logger=input.config.logger,
                destination_path=input.config.destination_path,
            )
        )

    async def execute(self) -> GetMediaFilesControllerOutput:
        try:
            _ = await self.use_case.execute()
            return GetMediaFilesControllerOutput(
                message="The extraction of media files was successfully!",
                status=HTTPStatus.OK.value,
            )
        except NetcameraConnectionError as error:
            return GetMediaFilesControllerOutput(
                message="An unnexpected error ocurred while extractingphotos and videos.",
                status=HTTPStatus.GATEWAY_TIMEOUT.value,
                metadata={"error": error.__str__()},
            )



from dataclasses import dataclass
from logging import Logger


@dataclass
class SaviiaAPIConfig:
    """
    Configuration for SAVIIA API.

    Attributes:
        ftp_port (int): Port number of the FTP server.
        ftp_host (str): Hostname or IP address of the FTP server.
        ftp_user (str): Username for the FTP server.
        ftp_password (str): Password for the FTP server.
        sharepoint_client_id (str): Client ID for SharePoint authentication.
        sharepoint_client_secret (str): Client secret for SharePoint authentication.
        sharepoint_tenant_id (str): Tenant ID for SharePoint authentication.
        sharepoint_tenant_name (str): Tenant name for SharePoint.
        sharepoint_site_name (str): Site name in SharePoint.
    """

    ftp_host: str
    ftp_port: int
    ftp_user: str
    ftp_password: str
    sharepoint_client_id: str
    sharepoint_client_secret: str
    sharepoint_tenant_id: str
    sharepoint_tenant_name: str
    sharepoint_site_name: str
    logger: Logger


@dataclass
class FtpClientConfig:
    ftp_host: str
    ftp_port: int
    ftp_user: str
    ftp_password: str


@dataclass
class SharepointConfig:
    sharepoint_client_id: str
    sharepoint_client_secret: str
    sharepoint_tenant_id: str
    sharepoint_tenant_name: str
    sharepoint_site_name: str



from aioftp import Client
from aioftp.errors import StatusCodeError
from saviialib.libs.ftp_client.ftp_client_contract import (
    FTPClientContract,
)
from saviialib.libs.ftp_client.types.ftp_client_types import (
    FtpClientInitArgs,
    FtpListFilesArgs,
    FtpReadFileArgs,
)


class AioFTPClient(FTPClientContract):
    def __init__(self, args: FtpClientInitArgs) -> None:
        self.host = args.config.ftp_host
        self.port = args.config.ftp_port
        self.password = args.config.ftp_password
        self.user = args.config.ftp_user
        self.client = Client()

    async def _async_start(self) -> None:
        try:
            await self.client.connect(host=self.host, port=self.port)
        except OSError:
            raise ConnectionRefusedError(
                f"{self.host}:{self.port} isn't active. "
                "Please ensure the server is running and accessible."
            )
        try:
            await self.client.login(user=self.user, password=self.password)
        except StatusCodeError:
            raise ConnectionAbortedError(
                "Authentication failed. Please verify your credentials and try again."
            )

    async def list_files(self, args: FtpListFilesArgs) -> list[str]:
        try:
            await self._async_start()
            files = []
            async for path, info in self.client.list(args.path, recursive=False):
                files.append((path.name, int((info.get("size", 0)))))
            return files
        except StatusCodeError as error:
            raise ConnectionAbortedError(error)

    async def read_file(self, args: FtpReadFileArgs) -> bytes:
        await self._async_start()
        try:
            async with self.client.download_stream(args.file_path) as stream:  # type: ignore
                return await stream.read()
        except StatusCodeError as error:
            raise FileNotFoundError(f"File not found: {args.file_path}") from error



# Internal libraries
from saviialib.libs.log_client.types.log_client_types import LogStatus


def format_message(class_name: str, method_name: str, status: LogStatus) -> str:
    return f"{class_name}: {method_name}_{status.value}"



from typing import Literal, Dict
from enum import Enum
from dataclasses import dataclass, field


class LogStatus(Enum):
    STARTED = "started"
    SUCCESSFUL = "successful"
    FAILED = "failed"
    ERROR = "error"
    EARLY_RETURN = "early_return"
    ALERT = "alert"


@dataclass
class LogClientArgs:
    client_name: str = field(default="logging")
    service_name: str = field(default="")
    class_name: str = field(default="")
    method_name: str = field(default="")
    active_record: bool = False


@dataclass
class InfoArgs:
    status: Literal[LogStatus.STARTED, LogStatus.SUCCESSFUL, LogStatus.EARLY_RETURN]
    metadata: Dict = field(default_factory=dict)


@dataclass
class DebugArgs:
    status: Literal[
        LogStatus.STARTED, LogStatus.SUCCESSFUL, LogStatus.EARLY_RETURN, LogStatus.ALERT
    ]
    metadata: Dict = field(default_factory=dict)


@dataclass
class ErrorArgs:
    status: Literal[LogStatus.ERROR]
    metadata: Dict = field(default_factory=dict)


@dataclass
class WarningArgs:
    status: Literal[LogStatus.FAILED, LogStatus.ALERT]
    metadata: Dict = field(default_factory=dict)



# Internal modules
from saviialib.libs.log_client.log_client_contract import LogClientContract
from saviialib.libs.log_client.logging_client.logging_client import LoggingClient
from saviialib.libs.log_client.types.log_client_types import (
    ErrorArgs,
    InfoArgs,
    LogClientArgs,
    DebugArgs,
    WarningArgs,
)


class LogClient(LogClientContract):
    def __init__(self, args: LogClientArgs):
        clients = {
            "logging": LoggingClient(args),
        }
        if clients.get(args.client_name):
            self.client_name = args.client_name
            self.client_obj = clients[args.client_name]
        else:
            raise KeyError(f"Unsupported client '{args.client_name}'.")

    @property
    def method_name(self):
        return self.client_obj.method_name

    @method_name.setter
    def method_name(self, method_name: str):
        self.client_obj.method_name = method_name

    @property
    def log_history(self):
        return self.client_obj.log_history

    def info(self, args: InfoArgs):
        return self.client_obj.info(args)

    def error(self, args: ErrorArgs):
        return self.client_obj.error(args)

    def debug(self, args: DebugArgs):
        return self.client_obj.debug(args)

    def warning(self, args: WarningArgs):
        return self.client_obj.warning(args)



from dataclasses import dataclass
from typing import Any


@dataclass
class FtpClientInitArgs:
    config: Any
    client_name: str = "aioftp_client"


@dataclass
class FtpListFilesArgs:
    path: str


@dataclass
class FtpReadFileArgs:
    file_path: str



from ....thies.controllers.types.update_thies_data_types import (
    UpdateThiesDataControllerInput,
    UpdateThiesDataControllerOutput,
)

__all__ = ["UpdateThiesDataControllerInput", "UpdateThiesDataControllerOutput"]



from http import HTTPStatus

from saviialib.general_types.error_types.api.saviia_api_error_types import (
    SharePointFetchingError,
    ThiesConnectionError,
    ThiesFetchingError,
    SharePointUploadError,
    SharePointDirectoryError,
)
from saviialib.general_types.error_types.common.common_types import (
    EmptyDataError,
    FtpClientError,
    SharepointClientError,
)
from saviialib.services.thies.controllers.types.update_thies_data_types import (
    UpdateThiesDataControllerInput,
    UpdateThiesDataControllerOutput,
)
from saviialib.services.backup.use_cases.types import (
    UpdateThiesDataUseCaseInput,
    SharepointConfig,
    FtpClientConfig,
)
from saviialib.services.thies.use_cases.update_thies_data import (
    UpdateThiesDataUseCase,
)


class UpdateThiesDataController:
    def __init__(self, input: UpdateThiesDataControllerInput):
        self.use_case = UpdateThiesDataUseCase(
            UpdateThiesDataUseCaseInput(
                ftp_config=FtpClientConfig(
                    ftp_host=input.config.ftp_host,
                    ftp_password=input.config.ftp_password,
                    ftp_port=input.config.ftp_port,
                    ftp_user=input.config.ftp_user,
                ),
                sharepoint_config=SharepointConfig(
                    sharepoint_client_id=input.config.sharepoint_client_id,
                    sharepoint_client_secret=input.config.sharepoint_client_secret,
                    sharepoint_site_name=input.config.sharepoint_site_name,
                    sharepoint_tenant_name=input.config.sharepoint_tenant_name,
                    sharepoint_tenant_id=input.config.sharepoint_tenant_id,
                ),
                sharepoint_folders_path=input.sharepoint_folders_path,
                ftp_server_folders_path=input.ftp_server_folders_path,
                local_backup_source_path=input.local_backup_source_path,
                logger=input.config.logger,
            )
        )

    async def execute(self) -> UpdateThiesDataControllerOutput:
        try:
            data = await self.use_case.execute()
            return UpdateThiesDataControllerOutput(
                message="THIES was synced successfully",
                status=HTTPStatus.OK.value,
                metadata={"data": data},  # type: ignore
            )
        except EmptyDataError:
            return UpdateThiesDataControllerOutput(
                message="No files to upload", status=HTTPStatus.NO_CONTENT.value
            )

        except (AttributeError, NameError, ValueError) as error:
            return UpdateThiesDataControllerOutput(
                message="An unexpected error occurred during use case initialization.",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )
        except FtpClientError as error:
            return UpdateThiesDataControllerOutput(
                message="Ftp Client initialization fails.",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )

        except SharepointClientError as error:
            return UpdateThiesDataControllerOutput(
                message="Sharepoint Client initialization fails.",
                status=HTTPStatus.INTERNAL_SERVER_ERROR.value,
                metadata={"error": error.__str__()},
            )

        except SharePointFetchingError as error:
            return UpdateThiesDataControllerOutput(
                message="An error occurred while retrieving file names from Microsoft SharePoint",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )

        except SharePointUploadError as error:
            return UpdateThiesDataControllerOutput(
                message="An error ocurred while uploading files to RCER Cloud",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )

        except SharePointDirectoryError as error:
            return UpdateThiesDataControllerOutput(
                message="An error ocurred while extracting folders from Microsoft Sharepoint",
                status=HTTPStatus.BAD_REQUEST.value,
                metadata={"error": error.__str__()},
            )

        except ThiesFetchingError as error:
            return UpdateThiesDataControllerOutput(
                message="An error ocurred while retrieving file names from THIES FTP Server.",
                status=HTTPStatus.NO_CONTENT.value,
                metadata={"error": error.__str__()},
            )

        except ThiesConnectionError as error:
            return UpdateThiesDataControllerOutput(
                message="Unable to connect to THIES Data Logger FTP Server.",
                status=HTTPStatus.INTERNAL_SERVER_ERROR.value,
                metadata={"error": error.__str__()},
            )



# External libraries
from abc import ABC, abstractmethod

# Internal modules
from saviialib.libs.log_client.types.log_client_types import (
    ErrorArgs,
    InfoArgs,
    DebugArgs,
    WarningArgs,
)


class LogClientContract(ABC):
    @abstractmethod
    def info(self, args: InfoArgs):
        pass

    @abstractmethod
    def error(self, args: ErrorArgs):
        pass

    @abstractmethod
    def debug(self, args: DebugArgs):
        pass

    @abstractmethod
    def warning(self, args: WarningArgs):
        pass



from saviialib.libs.sftp_client.sftp_client_contract import SFTPClientContract
from saviialib.libs.sftp_client.types.sftp_client_types import (
    SFTPClientInitArgs,
    ListfilesArgs,
    DownloadfilesArgs,
)
from typing import Optional, List, Tuple
from saviialib.general_types.error_types.common.common_types import EmptyDataError
import asyncssh  # type: ignore
from saviialib.libs.directory_client import DirectoryClient, DirectoryClientArgs


class AsyncsshSFTPClient(SFTPClientContract):
    def __init__(self, args: SFTPClientInitArgs) -> None:
        self.host: str = args.host
        self.port: int = args.port
        self.username: str = args.username
        self.password: Optional[str] = args.password
        self.ssh_key_path: Optional[str] = args.ssh_key_path
        self.dir_client = DirectoryClient(DirectoryClientArgs("os_client"))
        self._validate_credentials()

    def _validate_credentials(self):
        if not self.password and not self.ssh_key_path:
            raise EmptyDataError(
                reason="At least one attribute (ssh key or password) must be provided"
            )

    async def _start_connection(
        self,
    ) -> Tuple[asyncssh.SSHClientConnection, asyncssh.SFTPClient]:
        try:
            ssh_connection = await asyncssh.connect(
                self.host,
                username=self.username,
                port=self.port,
                client_keys=[self.ssh_key_path],
                password=self.password,
            )
            sftp_client = await ssh_connection.start_sftp_client()
            return ssh_connection, sftp_client
        except (OSError, asyncssh.Error) as exc:
            raise ConnectionError("SFTP Operation failed: " + str(exc))

    async def list_files(self, args: ListfilesArgs) -> List[str]:
        ssh_conn, sftp_client = await self._start_connection()
        async with ssh_conn:
            async with sftp_client:
                files = await sftp_client.listdir(args.path)
                files = [f for f in files if f not in (".", "..")]
                return files
        await self._end_connection(ssh_conn)

    async def download_files(self, args: DownloadfilesArgs) -> None:
        ssh_conn, sftp_client = await self._start_connection()
        if not args.destination_path or not args.source_path:
            conflict_path = (
                "destination path" if not args.destination_path else "source path"
            )
            raise ConnectionError(f"The {conflict_path} must be provided.")

        download_all = len(args.files_to_download) == 0
        async with ssh_conn:
            async with sftp_client:
                if download_all:
                    await sftp_client.get(
                        args.source_path,
                        args.destination_path,
                        recurse=True,
                        preserve=True,
                    )
                else:
                    for filename in args.files_to_download:
                        source_path = self.dir_client.join_paths(
                            args.source_path, filename
                        )
                        dest_path = self.dir_client.join_paths(
                            args.destination_path, filename
                        )
                        await sftp_client.get(source_path, dest_path)

    async def _end_connection(self, connection) -> None:
        await connection.wait_closed()



import aiofiles
from saviialib.libs.directory_client.directory_client import (
    DirectoryClient,
    DirectoryClientArgs,
)
from saviialib.libs.files_client.files_client_contract import FilesClientContract
from saviialib.libs.files_client.types.files_client_types import (
    FilesClientInitArgs,
    ReadArgs,
    WriteArgs,
)
import json


class AioFilesClient(FilesClientContract):
    def __init__(self, args: FilesClientInitArgs):
        self.dir_client = DirectoryClient(DirectoryClientArgs(client_name="os_client"))

    async def read(self, args: ReadArgs) -> str | bytes:
        if args.mode == "json":
            async with aiofiles.open(
                args.file_path, "r", encoding=args.encoding or "utf-8"
            ) as file:
                content = await file.read()
                return json.loads(content)

        encoding = None if args.mode == "rb" else args.encoding
        async with aiofiles.open(args.file_path, args.mode, encoding=encoding) as file:
            return await file.read()

    async def write(self, args: WriteArgs) -> None:
        file_path = (
            self.dir_client.join_paths(args.destination_path, args.file_name)
            if args.destination_path
            else args.file_name
        )
        if args.mode == "json":
            async with aiofiles.open(file_path, "w", encoding="utf-8") as file:
                json_str = json.dumps(args.file_content, ensure_ascii=False, indent=2)
                await file.write(json_str)
            return

        async with aiofiles.open(file_path, args.mode) as file:
            await file.write(args.file_content)



from dataclasses import dataclass
from logging import Logger
from typing import Dict, Tuple


@dataclass
class GetMediaFilesUseCaseInput:
    cameras: Dict[str, Tuple[str, int]]
    username: str
    password: str
    protocol: str
    logger: Logger
    destination_path: str


@dataclass
class GetMediaFilesUseCaseOutput:
    pass



from typing import Any
import json
from aiohttp import ClientError, ClientSession, TCPConnector
from dotenv import load_dotenv
import ssl
import certifi

from saviialib.libs.sharepoint_client.sharepoint_client_contract import (
    SharepointClientContract,
)
from saviialib.libs.sharepoint_client.types.sharepoint_client_types import (
    SpListFilesArgs,
    SpListFoldersArgs,
    SpUploadFileArgs,
    SpCreateFolderArgs,
    SharepointClientInitArgs,
)

load_dotenv()
ssl_context = ssl.create_default_context(cafile=certifi.where())


class SharepointRestAPI(SharepointClientContract):
    def __init__(self, args: SharepointClientInitArgs):
        self.session: ClientSession | None = None
        self.base_headers = {}
        self.credentials = {}
        self.base_url = ""
        self.tenant_id = args.config.sharepoint_tenant_id
        self.tenant_name = args.config.sharepoint_tenant_name
        self.client_secret = args.config.sharepoint_client_secret
        self.client_id = args.config.sharepoint_client_id
        self.site_name = args.config.sharepoint_site_name

    async def _load_form_digest_value(self) -> str:
        try:
            response = await self.session.post("contextinfo")
            response_json = await response.json()
            return response_json["FormDigestValue"]
        except ClientError as error:
            raise ConnectionError(error) from error

    async def _load_credentials(self) -> dict:
        resource_base = "00000003-0000-0ff1-ce00-000000000000"
        resource = f"{resource_base}/{self.tenant_name}.sharepoint.com@{self.tenant_id}"
        url = f"https://accounts.accesscontrol.windows.net/{self.tenant_id}/tokens/OAuth/2"
        payload = {
            "grant_type": "client_credentials",
            "client_id": f"{self.client_id}@{self.tenant_id}",
            "client_secret": self.client_secret,
            "resource": resource,
        }
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }

        async with ClientSession(connector=TCPConnector(ssl=ssl_context)) as session:
            # Load access token
            response = await session.post(url, data=payload, headers=headers)
            if response.status != 200:
                raise ClientError(
                    f"Failed to fetch credentials: {response.status}, {await response.text()}"
                )
            response_json = await response.json()

            return {
                "access_token": response_json["access_token"],
            }

    async def __aenter__(self) -> "SharepointRestAPI":
        try:
            self.credentials = await self._load_credentials()
            site_url = f"https://{self.tenant_name}.sharepoint.com"

            self.base_headers = {
                "Authorization": f"Bearer {self.credentials['access_token']}",
                "Accept": "application/json",
                "Content-Type": "application/json",
            }
            self.base_url = f"{site_url}/sites/{self.site_name}/_api/"
            connector = TCPConnector(ssl=ssl_context)
            self.session = ClientSession(
                headers=self.base_headers, base_url=self.base_url, connector=connector
            )
            return self
        except ClientError as error:
            raise ConnectionError(error)

    async def __aexit__(
        self, _exc_type: type[BaseException], _exc_val: BaseException, _exc_tb: Any
    ) -> None:
        await self.session.close()

    async def list_files(self, args: SpListFilesArgs) -> list:
        try:
            folder_relative_url = (
                f"GetFolderByServerRelativeUrl('{args.folder_relative_url}')"
            )
            endpoint = f"web/{folder_relative_url}/Files"
            response = await self.session.get(endpoint.lstrip("/"))
            response.raise_for_status()
            response_json = await response.json()
            return response_json
        except ClientError as error:
            raise ConnectionError(error) from error

    async def list_folders(self, args: SpListFoldersArgs) -> list:
        try:
            folder_relative_url = (
                f"GetFolderByServerRelativeUrl('{args.folder_relative_url}')"
            )
            endpoint = f"web/{folder_relative_url}/Folders"
            response = await self.session.get(endpoint.lstrip("/"))
            response.raise_for_status()
            response_json = await response.json()
            return response_json
        except ClientError as error:
            raise ConnectionError(error) from error

    async def upload_file(self, args: SpUploadFileArgs) -> dict:
        try:
            # Load form digest value
            form_digest_value = await self._load_form_digest_value()
            headers = {
                **self.base_headers,
                "X-RequestDigest": form_digest_value,
                "Content-Type": "application/octet-stream",
            }
            # Upload the file in the requested folder
            folder_relative_url = (
                f"GetFolderByServerRelativeUrl('{args.folder_relative_url}')"
            )
            data = args.file_content

            endpoint = f"web/{folder_relative_url}/Files/add(url='{args.file_name}',overwrite=true)"
            response = await self.session.post(endpoint, data=data, headers=headers)

            response.raise_for_status()
            return await response.json()
        except ClientError as error:
            raise ConnectionError(error) from error

    async def create_folder(self, args: SpCreateFolderArgs):
        try:
            # Load form digest value
            form_digest_value = await self._load_form_digest_value()
            headers = {
                **self.base_headers,
                "X-RequestDigest": form_digest_value,
            }
            body = {"ServerRelativeUrl": f"{args.folder_relative_url}"}
            endpoint = "web/folders"
            response = await self.session.post(
                endpoint.lstrip("/"), data=json.dumps(body), headers=headers
            )
            response.raise_for_status()
            response_json = await response.json()
            return response_json
        except ClientError as error:
            raise ConnectionError(error) from error



from dataclasses import dataclass, field
from typing import Dict

from saviialib.general_types.api.saviia_backup_api_types import SaviiaBackupConfig


@dataclass
class UploadBackupToSharepointControllerInput:
    config: SaviiaBackupConfig
    local_backup_source_path: str
    sharepoint_destination_path: str


@dataclass
class UploadBackupToSharepointControllerOutput:
    message: str
    status: int
    metadata: Dict[str, str] = field(default_factory=dict)



import asyncio
from .types.get_miniseed_files_types import (
    GetMiniseedFilesUseCaseInput,
    GetMiniseedFilesUseCaseOutput,
)
from typing import Dict, Any
from saviialib.libs.sftp_client import (
    SFTPClient,
    SFTPClientInitArgs,
    ListfilesArgs,
    DownloadfilesArgs,
)
from saviialib.libs.directory_client import DirectoryClient, DirectoryClientArgs
from saviialib.general_types.error_types.api.saviia_api_error_types import (
    ShakesNoContentError,
)
from saviialib.general_types.error_types.common.common_types import SftpClientError
from .utils.get_miniseed_files_utils import parse_downloaded_metadata


class GetMiniseedFilesUseCase:
    def __init__(self, input: GetMiniseedFilesUseCaseInput) -> None:
        self.password = input.password
        self.username = input.username
        self.ssh_key_path = input.ssh_key_path
        self.port = input.port
        self.raspberry_shakes: Dict[str, str] = input.raspberry_shakes
        self.dir_client = DirectoryClient(DirectoryClientArgs("os_client"))

    def _initialize_sftp_client(self, ip_address: str):
        return SFTPClient(
            SFTPClientInitArgs(
                "asyncssh_sftp",
                password=self.password,
                username=self.username,
                ssh_key_path=self.ssh_key_path,
                host=ip_address,
                port=self.port,
            )
        )

    async def _download_mseed_file(self, rs_name: str, rs_ip: str) -> Dict[str, Any]:
        DEST_BASE_DIR = "./rshakes-mseed-files"
        SOURCE_PATH = self.dir_client.join_paths("opt", "data", "archive")
        local_path = self.dir_client.join_paths(DEST_BASE_DIR, rs_name)
        if not await self.dir_client.isdir(local_path):
            await self.dir_client.makedirs(local_path)
        sftp_client = self._initialize_sftp_client(rs_ip)

        local_files = await self.dir_client.listdir(local_path)
        try:
            sftp_files = await sftp_client.list_files(ListfilesArgs(path=SOURCE_PATH))
            pending_files = set(sftp_files) - set(local_files)
            if not pending_files:
                raise ShakesNoContentError

            await sftp_client.download_files(
                DownloadfilesArgs(
                    source_path=SOURCE_PATH,
                    destination_path=local_path,
                    files_to_download=list(pending_files),
                )
            )
        except ConnectionError as error:
            raise SftpClientError(reason=error)
        return {
            "rs_name": rs_name,
            "destination_path": local_path,
            "total_files": len(pending_files),
        }

    async def execute(self):
        requests = []
        for rs_name, rs_ip in self.raspberry_shakes:
            requests.append(self._download_mseed_file(rs_name, rs_ip))
        responses = await asyncio.gather(*requests, return_exceptions=True)
        return GetMiniseedFilesUseCaseOutput(
            download_status=parse_downloaded_metadata(responses),
        )



from dataclasses import dataclass, field
from typing import Dict, Tuple
from saviialib.general_types.api.saviia_netcamera_api_types import SaviiaNetcameraConfig


@dataclass
class GetMediaFilesControllerInput:
    config = SaviiaNetcameraConfig
    cameras: Dict[str, Tuple[str, int]]


@dataclass
class GetMediaFilesControllerOutput:
    status: int
    message: str
    metadata: Dict = field(default_factory=dict)



from dataclasses import dataclass
from typing import Dict
from saviialib.general_types.api.saviia_shakes_api_types import SaviiaShakesConfig


@dataclass
class GetMiniseedFilesControllerInput:
    raspberry_shakes: Dict[str, str]
    config: SaviiaShakesConfig


@dataclass
class GetMiniseedFilesControllerOutput:
    status: int
    metadata: Dict
    message: str



# read version from installed package
from importlib.metadata import version

__version__ = version("saviialib")

from .general_types.api.saviia_api_types import SaviiaAPIConfig

from typing import Dict, Type, Any, overload, Literal, List

from saviialib.services.backup.api import SaviiaBackupAPI
from saviialib.services.thies.api import SaviiaThiesAPI
from saviialib.general_types.api.saviia_thies_api_types import SaviiaThiesConfig
from saviialib.general_types.api.saviia_backup_api_types import SaviiaBackupConfig

__all__ = ["SaviiaAPI", "SaviiaAPIConfig"]


class SaviiaAPI:
    API_REGISTRY: Dict[str, Type] = {
        "thies": SaviiaThiesAPI,
        "backup": SaviiaBackupAPI,
    }

    @overload
    def get(self, name: Literal["thies"]) -> SaviiaThiesAPI: ...
    @overload
    def get(self, name: Literal["backup"]) -> SaviiaBackupAPI: ...

    def __init__(self, config: SaviiaAPIConfig):
        """
        Receive a dictionary of configurations, with the same key
        as those registered in API_REGISTRY.

        :params configs: Dictionary of configurations for each API.

        Example:
            configs = {
                "thies": SaviiaThiesConfig(...),
                "backup": SaviiaBackupConfig(...)
            }
        """
        self._instances: Dict[str, Any] = {}

        for name, api_class in SaviiaAPI.API_REGISTRY.items():
            if name == "thies":
                service_config = SaviiaThiesConfig(
                    ftp_host=config.ftp_host,
                    ftp_port=config.ftp_port,
                    ftp_user=config.ftp_user,
                    ftp_password=config.ftp_password,
                    sharepoint_client_id=config.sharepoint_client_id,
                    sharepoint_client_secret=config.sharepoint_client_secret,
                    sharepoint_tenant_id=config.sharepoint_tenant_id,
                    sharepoint_tenant_name=config.sharepoint_tenant_name,
                    sharepoint_site_name=config.sharepoint_site_name,
                    logger=config.logger,
                )
            elif name == "backup":
                service_config = SaviiaBackupConfig(
                    sharepoint_client_id=config.sharepoint_client_id,
                    sharepoint_client_secret=config.sharepoint_client_secret,
                    sharepoint_tenant_id=config.sharepoint_tenant_id,
                    sharepoint_tenant_name=config.sharepoint_tenant_name,
                    sharepoint_site_name=config.sharepoint_site_name,
                    logger=config.logger,
                )

            self._instances[name] = api_class(service_config)

    def get(self, name: Literal["thies", "backup"]) -> Any:
        """Returns the API instance associated with the given name."""
        try:
            return self._instances[name]
        except KeyError:
            raise ValueError(f"API '{name}' is not registered or not configured.")

    def list_available(self) -> List[str]:
        """List of available registered APIs."""
        return list(self._instances.keys())



from ....thies.use_cases.types.update_thies_data_types import (
    FtpClientConfig,
    SharepointConfig,
    UpdateThiesDataUseCaseInput,
)

__all__ = ["UpdateThiesDataUseCaseInput", "FtpClientConfig", "SharepointConfig"]



from typing import Dict, List, Any


def parse_downloaded_metadata(responses: List[Any]) -> Dict[str, Any]:
    return {
        x["rs_name"]: {
            "destination_path": x["destination_path"],
            "total_files": x["total_files"],
        }
        for x in responses
    }



# External libraries
import logging

# Internal modules
from saviialib.libs.log_client.utils.log_client_utils import format_message
from saviialib.libs.log_client.log_client_contract import LogClientContract
from saviialib.libs.log_client.types.log_client_types import (
    ErrorArgs,
    InfoArgs,
    LogClientArgs,
    DebugArgs,
    WarningArgs,
)

from saviialib.libs.zero_dependency.utils.datetime_utils import today, datetime_to_str


class LoggingClient(LogClientContract):
    def __init__(self, args: LogClientArgs):
        log_format = "{message}"
        logging.basicConfig(format=log_format, level=logging.INFO, style="{")
        self.class_name = args.class_name
        self.method_name = args.method_name
        self.active_record = args.active_record
        self.log_history = []

    def _save_to_history(self, meta: dict):
        self.log_history.append(
            f"[{datetime_to_str(today(), date_format='%m-%d-%Y %H:%M:%S')}][{self.class_name}] {meta.get('msg', '')}"
        )

    def info(self, args: InfoArgs) -> None:
        if self.active_record:
            self._save_to_history(args.metadata)
        return logging.info(
            format_message(self.class_name, self.method_name, args.status)
        )

    def error(self, args: ErrorArgs) -> None:
        if self.active_record:
            self._save_to_history(args.metadata)
        return logging.error(
            format_message(self.class_name, self.method_name, args.status)
        )

    def debug(self, args: DebugArgs) -> None:
        if self.active_record:
            self._save_to_history(args.metadata)
        return logging.debug(
            format_message(self.class_name, self.method_name, args.status)
        )

    def warning(self, args: WarningArgs) -> None:
        if self.active_record:
            self._save_to_history(args.metadata)
        return logging.warning(
            format_message(self.class_name, self.method_name, args.status)
        )



from saviialib.libs.directory_client.directory_client_contract import (
    DirectoryClientContract,
)
import os
import asyncio


class OsClient(DirectoryClientContract):
    @staticmethod
    def join_paths(*paths: str) -> str:
        return os.path.join(*paths)

    @staticmethod
    async def path_exists(path: str) -> bool:
        return await asyncio.to_thread(os.path.exists, path)

    @staticmethod
    async def listdir(path: str, more_info: bool = False) -> list:
        def _listdir_with_size(path):
            items = []
            for name in os.listdir(path):
                full_path = os.path.join(path, name)
                is_dir = os.path.isdir(full_path)
                size = os.stat(full_path).st_size if not is_dir else 0
                items.append((name, size))
            return items

        if more_info:
            return await asyncio.to_thread(_listdir_with_size, path)
        return await asyncio.to_thread(os.listdir, path)

    @staticmethod
    async def isdir(path: str) -> bool:
        return await asyncio.to_thread(os.path.isdir, path)

    @staticmethod
    async def makedirs(path: str) -> None:
        return await asyncio.to_thread(os.makedirs, path, exist_ok=True)

    @staticmethod
    async def remove_file(path: str) -> None:
        if await asyncio.to_thread(os.path.exists, path):
            await asyncio.to_thread(os.remove, path)

    @staticmethod
    async def walk(path: str):
        return await asyncio.to_thread(os.walk, path)

    @staticmethod
    def relative_path(full_path: str, base_folder: str):
        return os.path.relpath(full_path, base_folder)

    @staticmethod
    def get_basename(path: str):
        return os.path.basename(path)



from .types.get_media_files_types import (
    GetMediaFilesUseCaseInput,
    GetMediaFilesUseCaseOutput,
)
from saviialib.libs.ffmpeg_client import (
    FfmpegClient,
    FfmpegClientInitArgs,
    RecordVideoArgs,
    RecordPhotoArgs,
)
from saviialib.libs.zero_dependency.utils.strings_utils import are_equal
from typing import Tuple, Dict
from saviialib.libs.directory_client.directory_client import (
    DirectoryClient,
    DirectoryClientArgs,
)
from saviialib.general_types.error_types.api.saviia_netcamera_error_types import (
    NetcameraConnectionError,
)


class GetMediaFilesUseCase:
    def __init__(self, input: GetMediaFilesUseCaseInput) -> None:
        self.ffmpeg_client = FfmpegClient(
            FfmpegClientInitArgs(
                client_name="ffmpeg_asyncio",
            )
        )
        self.dir_client = DirectoryClient(DirectoryClientArgs("os_client"))
        self.user = input.username
        self.pwd = input.password
        self.cameras: Dict[str, Tuple[str, int]] = input.cameras
        self.protocol = input.protocol
        self.logger = input.logger
        self.dest_path = input.destination_path

    async def _retieve_with_rtsp(self):
        for name, conn in self.cameras.items():
            ip, port = conn
            dest_path = self.dir_client.join_paths(self.dest_path, name)
            try:
                # Extraction of photo files into dest_path dir.
                await self.ffmpeg_client.record_photo(
                    RecordPhotoArgs(
                        ip_address=ip,
                        port=str(port),
                        destination_path=dest_path,
                        rtsp_user=self.user,
                        rtsp_password=self.pwd,
                        extension="jpg",
                        frames=1,
                    )
                )
                # Extraction of video files into dest_path dir.
                await self.ffmpeg_client.record_video(
                    RecordVideoArgs(
                        destination_path=dest_path,
                        ip_address=ip,
                        port=str(port),
                        rtsp_user=self.user,
                        rtsp_password=self.pwd,
                        extension="mp3",
                        duration=10,
                    )
                )
            except ConnectionError as error:
                raise NetcameraConnectionError(reason=error)

    async def execute(self) -> GetMediaFilesUseCaseOutput:
        if are_equal(self.protocol, "rtsp"):
            await self._retieve_with_rtsp()
        else:
            raise NotImplementedError(
                f"The media files extraction with {self.protocol} is not implemented yet."
            )
        return GetMediaFilesUseCaseOutput()



from typing import Any, Dict

from .controllers.upload_backup_to_sharepoint import (
    UploadBackupToSharepointControllerInput,
)
from .controllers.upload_backup_to_sharepoint import UploadBackupToSharepointController
from saviialib.general_types.api.saviia_backup_api_types import SaviiaBackupConfig


class SaviiaBackupAPI:
    """
    EpiiAPI is a service class that provides methods to interact with Patagonia Center system.
    """

    def __init__(self, config: SaviiaBackupConfig):
        self.config = config

    async def upload_backup_to_sharepoint(
        self, local_backup_source_path: str, sharepoint_destination_path: str
    ) -> Dict[str, Any]:
        """Migrate a backup folder from Home assistant to Sharepoint directory.
        Args:
            local_backup_source_path (str): Local path to backup.
        Returns:
            response (dict): A dictionary containing the response from the upload operation.
                This dictionary will typically include information about the success or
                failure of the upload, as well as any relevant metadata.
        """

        controller = UploadBackupToSharepointController(
            UploadBackupToSharepointControllerInput(
                self.config, local_backup_source_path, sharepoint_destination_path
            )
        )
        response = await controller.execute()
        return response.__dict__



from dataclasses import dataclass
from logging import Logger


@dataclass
class SaviiaShakesConfig:
    """
    Configuration for Raspberry shakes activities as Miniseed extraction, photo record and video record.

    Attributes:
        sftp_user (str): Username for SFTP Client connection
        sftp_password (str): Password for SFTP Client connection
        sftp_port (str): SFTP Server Port. Default port is 22.
        ssh_key_path (str): Path to the SSH Private key for client-side authentication.
    """

    sftp_user: str
    sftp_password: str
    ssh_key_path: str
    logger: Logger
    sftp_port: int = 22



from dataclasses import dataclass, field
from typing import Dict
from saviialib.general_types.api.saviia_thies_api_types import SaviiaThiesConfig


@dataclass
class UpdateThiesDataControllerInput:
    config: SaviiaThiesConfig
    sharepoint_folders_path: list
    ftp_server_folders_path: list
    local_backup_source_path: str


@dataclass
class UpdateThiesDataControllerOutput:
    message: str
    status: int
    metadata: Dict[str, str] = field(default_factory=dict)



class NetcameraConnectionError(Exception):
    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return "Netcamera Connection failed. " + self.reason.__str__()



from .controllers.get_media_files import (
    GetMediaFilesController,
    GetMediaFilesControllerInput,
)
from typing import Dict, Tuple, Any
from saviialib.general_types.api.saviia_netcamera_api_types import SaviiaNetcameraConfig


class NetcameraAPI:
    """This class provides methods for interacting with network cameras and retrieving
    files using streaming services."""

    def __init__(self, config: SaviiaNetcameraConfig):
        self.config = config

    async def get_media_files(
        self, cameras: Dict[str, Tuple[str, int]]
    ) -> Dict[str, Any]:
        """Retrieve media files from Network cameras.

        :param cameras: Dictionary where the key is the identifier of the camera, and the
            value is a tuple wich contains the service IP address and port of connection.
            Example: {'cam_01': ('192.168.1.10', 8080), ...}
        :type cameras: dict
        :return response: A dictionary containg information of the extraction operation.
        :rtype: dict
        """
        controller = GetMediaFilesController(GetMediaFilesControllerInput(cameras))
        response = await controller.execute()
        return response.__dict__



from .thies_bp import THIESDayData
import pandas as pd
from logging import Logger
from asyncio import to_thread
from saviialib.libs.directory_client import DirectoryClient
from saviialib.libs.zero_dependency.utils.datetime_utils import datetime_to_str, today
from saviialib.libs.files_client import FilesClient, FilesClientInitArgs, WriteArgs
import saviialib.services.thies.constants.update_thies_data_constants as c


async def create_thies_daily_statistics_file(
    local_backup_path: str, os_client: DirectoryClient, logger: Logger
) -> None:
    csv_client = FilesClient(FilesClientInitArgs(client_name="csv_client"))
    filename = datetime_to_str(today(), date_format="%Y%m%d") + ".BIN"
    logger.debug(
        f"[thies_synchronization_lib] Creating Daily Statistics for {filename}"
    )
    path_bin_av = os_client.join_paths(local_backup_path, "thies", "AVG", filename)
    path_ini_av = os_client.join_paths(
        local_backup_path, "thies", "AVG", "DESCFILE.INI"
    )
    path_bin_ex = os_client.join_paths(local_backup_path, "thies", "EXT", filename)
    path_ini_ex = os_client.join_paths(
        local_backup_path, "thies", "EXT", "DESCFILE.INI"
    )

    ext_df = THIESDayData("ex")
    await to_thread(ext_df.read_binfile, path_bin_ex, path_ini_ex)

    avg_df = THIESDayData("av")
    await to_thread(avg_df.read_binfile, path_bin_av, path_ini_av)

    ext_df = ext_df.dataDF[c.EXT_COLUMNS.keys()]
    avg_df = avg_df.dataDF[c.AVG_COLUMNS.keys()]

    # Merge both dataframes
    df = avg_df.merge(ext_df, on=["Date", "Time"], how="outer")
    # Set the date as dd.mm.yyyy format.
    df["Date"] = df["Date"].str.replace(
        r"(\d{4})/(\d{2})/(\d{2})", r"\3.\2.\1", regex=True
    )
    df["Hour"] = df["Time"].str[:2]

    # Group by hour.
    hourly_agg = df.groupby(["Date", "Hour"]).agg(c.AGG_DICT).reset_index()

    rows = []
    # For each attribute in avg_columns (except Date, Time)
    for col, col_id in c.AVG_COLUMNS.items():
        if col in ["Date", "Time"]:
            continue
        # Determine the corresponding min/max columns if they exist
        min_col = f"{col} MIN"
        max_col = f"{col} MAX"
        mean_col = col
        if col in ["WS", "WD"]:
            max_col += " gust"

        unit = c.UNITS.get(col, "")

        for idx, row in hourly_agg.iterrows():
            statistic_id = f"sensor.saviia_epii_{col_id}"
            start = f"{row['Date']} {row['Hour']}:00"
            mean = row[mean_col] if mean_col in row else 0
            min_val = row[min_col] if min_col in row else mean
            max_val = row[max_col] if max_col in row else mean

            # If no min/max for this attribute, set as Na or 0 as requested
            if not (pd.isna(mean) or pd.isna(min_val) or pd.isna(max_val)):
                pass
            elif pd.isna(mean) and not (pd.isna(min_val) or pd.isna(max_val)):
                mean = (min_val + max_val) / 2
            else:
                val_notna = [x for x in {mean, min_val, max_val} if not pd.isna(x)]
                if len(val_notna) >= 1:
                    mean_val = sum(val_notna) / len(val_notna)
                    mean = max_val = min_val = mean_val
                else:
                    continue  # Do not consider a row with null data

            # Normalize if the mean is upper than maxval or lower than minval
            if (mean < min_val or mean > max_val) and col not in ["WD"]:
                mean = (min_val + max_val) / 2
            
            # Filter battery values if they are out of the range 
            if col_id == "battery":
                values = [mean, max_val, min_val]
                if any(v < 0 or v > 50 for v in values):
                    continue
                

            if col in ["WD"]:  # Avoid error
                rows.append(
                    {
                        "statistic_id": statistic_id,
                        "unit": unit,
                        "start": start,
                        "min": mean,
                        "max": mean,
                        "mean": mean,
                    }
                )
            else:
                rows.append(
                    {
                        "statistic_id": statistic_id,
                        "unit": unit,
                        "start": start,
                        "min": min_val,
                        "max": max_val,
                        "mean": mean,
                    }
                )

    logger.debug("[thies_synchronization_lib] Saving file in the main directory")
    await csv_client.write(
        WriteArgs(file_name="thies_daily_statistics.tsv", file_content=rows, mode="w")
    )
    logger.debug(
        "[thies_synchronization_lib] thies_daily_statistics.tsv created successfully!"
    )



from typing import Any, Dict, List

from .controllers.types.update_thies_data_types import UpdateThiesDataControllerInput
from .controllers.update_thies_data import UpdateThiesDataController
from saviialib.general_types.api.saviia_thies_api_types import (
    SaviiaThiesConfig,
)


class SaviiaThiesAPI:
    def __init__(self, config: SaviiaThiesConfig) -> None:
        self.config = config

    async def update_thies_data(
        self,
        sharepoint_folders_path: List[str],
        ftp_server_folders_path: List[str],
        local_backup_source_path: str,
    ) -> Dict[str, Any]:
        """Updates data from a THIES Data Logger by connecting to an FTP server
        and transferring data to specified Sharepoint folders.

        :param list sharepoint_folders_path: List of Sharepoint folder paths for AVG and EXT data.
            The AVG path must be the first element.
        :param list ftp_server_folders_path: List of FTP server folder paths for AVG and EXT data.
            The AVG path must be the first element.
        :param str local_backup_source_path: Path of the main directory where the files extracted from
            the Thies FTP Server are going to be stored

        :return: A dictionary representation of the API response.
        :rtype: dict
        """
        controller = UpdateThiesDataController(
            UpdateThiesDataControllerInput(
                self.config,
                sharepoint_folders_path,
                ftp_server_folders_path,
                local_backup_source_path,
            )
        )
        response = await controller.execute()
        return response.__dict__


from .sharepoint_client import SharepointClient
from .types.sharepoint_client_types import (
    SharepointClientInitArgs,
    SpListFilesArgs,
    SpListFoldersArgs,
    SpUploadFileArgs,
    SpCreateFolderArgs,
)

__all__ = [
    "SharepointClientInitArgs",
    "SharepointClient",
    "SpListFilesArgs",
    "SpListFoldersArgs",
    "SpUploadFileArgs",
    "SpCreateFolderArgs",
]



import json


class ThiesConnectionError(Exception):
    """Raised when unable to connect to the THIES FTP Server"""

    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return "Unable to connect to THIES FTP Server. " + self.reason.__str__()


class ThiesFetchingError(Exception):
    """Raised when no files are found to upload to the server."""

    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return (
            "An error ocurred while retrieving files from THIES FTP Server. "
            + self.reason.__str__()
        )


class SharePointFetchingError(Exception):
    """Raised when there is an error fetching file names from the RCER cloud."""

    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        try:
            _, internal_metadata = self.reason.__str__().split(",", 1)
            internal_metadata_dict = json.loads(internal_metadata)
            return internal_metadata_dict["error_description"]

        except json.decoder.JSONDecodeError:
            return self.reason.__str__()


class SharePointDirectoryError(Exception):
    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return (
            "An error occurred while fetching the folders from Microsoft SharePoint. "
            + self.reason.__str__()
        )


class SharePointUploadError(Exception):
    """Raised when there is an error uploading files to the Microsoft SharePoint folder."""

    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return (
            "An error occurred while uploading files to the Microsoft SharePoint folder. "
            + self.reason.__str__()
        )


class BackupUploadError(Exception):
    """Raised when there is an error when occurs the migration from the local backup to
    sharepoint cloud."""

    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return (
            "An error occurred during the migration from the local backup to SharePoint cloud. "
            "Search the logs for more information. " + self.reason.__str__()
        )


class BackupSourcePathError(Exception):
    """Raised when the local backup source path is invalid."""

    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return "Invalid local backup source path. " + self.reason.__str__()


class BackupEmptyError(Exception):
    """Raised when the local backup folder is empty."""

    def __init__(self, *args):
        super().__init__(*args)

    def __str__(self):
        return "The local backup folder is empty. "


class ShakesNoContentError(Exception):
    def __init__(self, *args):
        super().__init__(*args)

    def __str__(self):
        return "All the miniSEED files have been downloaded and are in the local directory."



AVG_COLUMNS = {
    "Date": "date",
    "Time": "time",
    "AirTemperature": "air_temperature",
    "Radiation": "radiation",
    "CO2": "carbon_dioxide",
    "Precipitation": "precipitation",
    "WS": "wind_velocity",
    "WD": "wind_direction",
    "Humidity": "humidity",
    "UBat": "battery",
}

EXT_COLUMNS = {
    "Date": "date",
    "Time": "time",
    "AirTemperature MIN": "air_temperature",
    "AirTemperature MAX": "air_temperature",
    "Radiation MIN": "radiation",
    "Radiation MAX": "radiation",
    "CO2 MIN": "carbon_dioxide",
    "CO2 MAX": "carbon_dioxide",
    "WS MIN": "wind_velocity",
    "WS MAX gust": "wind_velocity",
    "WD MIN": "wind_direction",
    "WD MAX gust": "wind_direction",
    "Humidity MIN": "humidity",
    "Humidity MAX": "humidity",
    "UBat MIN": "battery",
    "UBat MAX": "battery",
}

AGG_DICT = {
    "AirTemperature": "mean",
    "AirTemperature MIN": "mean",
    "AirTemperature MAX": "mean",
    "Precipitation": "sum",
    "Humidity": "mean",
    "Humidity MIN": "mean",
    "Humidity MAX": "mean",
    "Radiation": "sum",
    "Radiation MIN": "sum",
    "Radiation MAX": "sum",
    "CO2": "mean",
    "CO2 MIN": "mean",
    "CO2 MAX": "mean",
    "WS": "mean",
    "WS MIN": "mean",
    "WS MAX gust": "mean",
    "WD": "mean",
    "WD MIN": "mean",
    "WD MAX gust": "mean",
    "UBat": "mean",
    "UBat MIN": "mean",
    "UBat MAX": "mean",
}

UNITS = {
    "AirTemperature": "°C",
    "Precipitation": "mm",
    "Humidity": "%",
    "Radiation": "W/m²",
    "CO2": "ppm",
    "WS": "m/s",
    "WD": "°",
    "UBat": "V",
}



from dataclasses import dataclass
from logging import Logger


@dataclass
class SaviiaNetcameraConfig:
    username: str
    password: str
    protocol: str
    logger: Logger
    destination_path: str = "/"



from abc import ABC, abstractmethod
from typing import Iterator


class DirectoryClientContract(ABC):
    @abstractmethod
    def join_paths(self, *paths: str) -> str:
        pass

    @abstractmethod
    async def path_exists(self, path: str) -> bool:
        pass

    @abstractmethod
    async def listdir(self, path: str, more_info: bool = False) -> list:
        pass

    @abstractmethod
    async def isdir(self, path) -> bool:
        pass

    @abstractmethod
    async def makedirs(self, path: str) -> None:
        pass

    @abstractmethod
    async def remove_file(self, path: str) -> None:
        pass

    @abstractmethod
    async def walk(self, path: str) -> Iterator:
        pass

    @abstractmethod
    def relative_path(self, full_path: str, base_folder: str):
        pass

    @abstractmethod
    def get_basename(self, path: str):
        pass



import configparser
import os
import struct
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from bitarray import bitarray

TIME_LIST = []

start_time = datetime.strptime("00:00", "%H:%M")
for i in range(0, 24 * 60, 10):
    TIME_LIST.append((start_time + timedelta(minutes=i)).strftime("%H:%M"))

ROWS = len(TIME_LIST)


def date_range(start_date: str, end_date: str) -> list:
    start = datetime.strptime(start_date, "%Y/%m/%d") + timedelta(days=1)
    end = datetime.strptime(end_date, "%Y/%m/%d") - timedelta(days=1)
    return [
        (start + timedelta(days=i)).strftime("%Y/%m/%d")
        for i in range((end - start).days + 1)
        if start <= end
    ]


def add_date_sep(date: str) -> str:
    """
    Input: date as YYYYMMDD.BIN
    Returns: date as YYYY/MM/DD
    """
    return date[:4] + "/" + date[4:6] + "/" + date[6:8]


def verify_datestr(filename: str) -> bool:
    """
    Returns True if filename has the YYYYMMDD.BIN format
    """
    try:
        datetime.strptime(filename[:8], "%Y%m%d")
        return filename.endswith(".BIN")
    except ValueError:
        return False


def read_descfile(path) -> dict:
    """
    Input: path DESCFILE.INI
    Returns: dict
        key is index [i]
        value is dict with parameters from .ini
    """
    if type(path) == dict:  # noqa: E721
        return path
    config = configparser.ConfigParser()
    config.read(path)
    data_dict = {}
    for section in config.sections():
        section_dict = dict(config.items(section))
        for v in section_dict:
            if v == "name":
                continue
            section_dict[v] = int(section_dict[v])
        data_dict[int(section)] = section_dict
    return data_dict


class THIESDayData:
    # Bytes per parameter
    BPP = {"av": 5, "ex": 9}
    # Timestamp Offset
    OFFSET = 4

    def __init__(self, datatype: str) -> None:
        d = datatype.lower().strip()
        if d not in ["av", "ex"]:
            raise ValueError(
                "Invalid datatype. Expected 'av' (average values) or 'ex' (minmax values)."
            )

        self._bpr = -1  # Bytes per row
        self._bpp = THIESDayData.BPP[d]  # Bytes per parameter
        self._datatype = d
        self._binfile = None
        self.descfile = {}
        self.nparameters = -1
        self._parameters = []
        self.nbytes = -1
        self.nrows = -1
        self._date = ""
        self.statusDF = pd.DataFrame()
        self.dataDF = pd.DataFrame()
        self.datesDF = pd.DataFrame()

    @staticmethod
    def _bytes2datetime(b: bytes, only_time: bool = False) -> str:
        """
        Input: bytes (size 4)
        Output: str (YYYY/MM/DD hh:mm:ss)
        """
        bits = bitarray()
        bits.frombytes(b[::-1])  # Invert 4 bytes
        hr = int(bits[15:20].to01(), 2)
        min = int(bits[20:26].to01(), 2)
        sec = int(bits[26:].to01(), 2)
        time = f"{str(hr).zfill(2)}:{str(min).zfill(2)}"
        if only_time:
            return time
        yr = int(bits[0:6].to01(), 2)
        mon = int(bits[6:10].to01(), 2)
        day = int(bits[10:15].to01(), 2)
        date = f"20{yr}/{str(mon).zfill(2)}/{str(day).zfill(2)}"
        return date + " " + time + f":{str(sec).zfill(2)}"

    def _set_descfile(self, inipath: str) -> None:
        self.descfile = read_descfile(inipath)
        self.nparameters = len(self.descfile)
        row_size = sum([self.descfile[num]["size"] for num in self.descfile])
        self._bpr = row_size + THIESDayData.OFFSET

    def read_binfile(self, binpath: str, inipath: str) -> None:
        self._set_descfile(inipath)
        with open(binpath, "rb") as bin_file:
            binfile = bin_file.read()
        self._binfile = binfile
        self.nbytes = len(self._binfile)
        self.nrows = int(self.nbytes / self._bpr)
        self._make_dataframes()

    def make_empty(self, inipath: str, date: str) -> None:
        self._set_descfile(inipath)
        dataDF = pd.DataFrame(
            None, index=range(ROWS), columns=range(self.nparameters + 2)
        )
        col_names = {0: "Date", 1: "Time"}
        par_names = {key + 1: self.descfile[key]["name"] for key in self.descfile}
        col_names.update(par_names)
        dataDF = dataDF.rename(columns=col_names)
        dataDF["Time"] = TIME_LIST
        dataDF["Date"] = [date] * ROWS

        self.dataDF = dataDF
        self.statusDF = dataDF
        self.datesDF = dataDF

    def _make_dataframes(self) -> None:
        """
        Builds data DF, status DF and, if datatype=ex, dates DF.
        """
        byterows = [
            self._binfile[i * self._bpr + THIESDayData.OFFSET : (i + 1) * self._bpr]
            for i in range(0, self.nrows)
        ]
        data_arr = np.zeros((self.nrows, self.nparameters))
        status_arr = np.zeros((self.nrows, self.nparameters))
        time_idx = np.empty(self.nrows, dtype=object)
        date_idx = np.empty(self.nrows, dtype=object)
        dates_arr = np.empty((self.nrows, self.nparameters), dtype=object)

        for i, row in enumerate(byterows):
            # Timestamp
            ts_bytes = self._binfile[i * self._bpr : i * self._bpr + 4]
            ts = THIESDayData._bytes2datetime(ts_bytes)
            date_idx[i], time_idx[i] = ts[:-3].split()

            for j in range(self.nparameters):
                # Status = byte 1
                status = row[j * self._bpp]
                status_arr[i, j] = status

                # Value = bytes 2-5, float
                value = struct.unpack("<f", row[j * self._bpp + 1 : j * self._bpp + 5])[
                    0
                ]
                data_arr[i, j] = round(value, 1)

                if self._datatype == "ex":
                    # Datetime = bytes 6-9
                    dt = THIESDayData._bytes2datetime(
                        row[j * self._bpp + 5 : j * self._bpp + 9], only_time=True
                    )
                    dates_arr[i, j] = dt

        self.dataDF = pd.DataFrame(data_arr).rename(
            columns={i: self.descfile[i + 1]["name"] for i in range(self.nparameters)}
        )
        self.statusDF = pd.DataFrame(status_arr).rename(
            columns={i: self.descfile[i + 1]["name"] for i in range(self.nparameters)}
        )
        self.dataDF = self.dataDF.where(self.statusDF == 0.0, other=None)

        if self._datatype == "ex":
            self.datesDF = pd.DataFrame(dates_arr).rename(
                columns={
                    i: self.descfile[i + 1]["name"] for i in range(self.nparameters)
                }
            )
            self.datesDF = self.datesDF.where(self.statusDF == 0.0, other=None)
            self.datesDF.insert(0, "Time", time_idx)
            self.datesDF.insert(0, "Date", date_idx)

        self.dataDF.insert(0, "Time", time_idx)
        self.dataDF.insert(0, "Date", date_idx)
        self.statusDF.insert(0, "Time", time_idx)
        self.statusDF.insert(0, "Date", date_idx)

    def _generate_blank_rows(self) -> pd.DataFrame:
        if len(self) == ROWS:
            # Nothing to fill (already full rows)
            return []

        new = []
        none_row = {col: None for col in self.dataDF.columns}
        none_row["Date"] = self.date
        current_times = self.dataDF["Time"]
        for time in TIME_LIST:
            if time not in current_times.values:
                row = none_row.copy()
                # 'time' was not measured in the original data
                # fill it with None row
                row["Time"] = time
                new.append(row)
        return pd.DataFrame(new)

    def complete_empty(self):
        """
        Completes DataFrames with all the timestamps of missing data
        Fills all columns with 'None' except Date and Time cols
        """
        if len(self) == ROWS:
            return
        new_rows = self._generate_blank_rows()
        # self.dataDF = self.dataDF.append(new_rows, ignore_index=True)
        self.dataDF = pd.concat([self.dataDF, new_rows], ignore_index=True)
        self.dataDF = self.dataDF.sort_values(by="Time").reset_index(drop=True)
        # self.statusDF = self.statusDF.append(new_rows, ignore_index=True)
        self.statusDF = pd.concat([self.statusDF, new_rows], ignore_index=True)
        self.statusDF = self.statusDF.sort_values(by="Time").reset_index(drop=True)

        if self._datatype == "ex":
            # self.datesDF = self.datesDF.append(new_rows, ignore_index=True)
            self.datesDF = pd.concat([self.datesDF, new_rows], ignore_index=True)
            self.datesDF = self.datesDF.sort_values(by="Time").reset_index(drop=True)

    def sort_by(self, cols: list):
        self.dataDF = self.dataDF.sort_values(
            by=cols, ascending=[True, True]
        ).reset_index(drop=True)
        self.statusDF = self.statusDF.sort_values(
            by=cols, ascending=[True, True]
        ).reset_index(drop=True)
        if len(self.datesDF):
            self.datesDF = self.datesDF.sort_values(
                by=cols, ascending=[True, True]
            ).reset_index(drop=True)

    @property
    def date(self) -> str:
        """
        Returns str of date of measurement
        """
        if len(self.dataDF) and self._date == "":
            self._date = self.dataDF["Date"][0]
        return self._date

    @property
    def shape(self):
        return self.dataDF.shape

    @property
    def info(self) -> None:
        bf = self._binfile
        if bf:
            bf = bf[:8]
        print(f"""=== THIES Day Data Instance ===\n
Bytes per row (BPR): {self._bpr}
Bytes per parameter (BPP): {self._bpp}
Datatype: {self._datatype}
Binfile: {bf}...
Descfile: {self.descfile}
N parameters: {self.nparameters}
N Bytes: {self.nbytes}
Rows: {self.nrows}
Date: {self.date}
    """)

    @property
    def parameters(self) -> list:
        if self._parameters == []:
            self._parameters = [self.descfile[i]["name"] for i in self.descfile]
        return self._parameters

    def write_csv(self, filename: str) -> None:
        with open(filename + ".csv", "w") as outfile:
            outfile.write(self.dataDF.to_csv())

    def __repr__(self) -> str:
        return str(self.dataDF)

    def _repr_html_(self):
        return self.dataDF._repr_html_()

    def __len__(self):
        return len(self.dataDF)

    def __add__(self, other):
        if isinstance(other, THIESDayData):
            new = THIESDayData(datatype=self._datatype)
            new.descfile = other.descfile
            new.nparameters = other.nparameters
            new._parameters = other.parameters
            new.nrows = self.nrows + other.nrows
            new.nbytes = self.nbytes + other.nbytes
            new.statusDF = pd.concat([self.statusDF, other.statusDF]).reset_index(
                drop=True
            )
            new.dataDF = pd.concat([self.dataDF, other.dataDF]).reset_index(drop=True)
            if self._datatype == "ex":
                new.datesDF = pd.concat([self.datesDF, other.datesDF]).reset_index(
                    drop=True
                )
            return new
        raise TypeError(
            f"unsupported operand type(s) for +: 'THIESDayData' and '{type(other)}'"
        )


class THIESData:
    def __init__(self, datatype: str, dirpath: str) -> None:
        d = datatype.lower().strip()
        if d not in ["av", "ex"]:
            raise ValueError(
                "Invalid datatype. Expected 'av' (average values) or 'ex' (minmax values)."
            )

        self._path = dirpath
        self._datatype = d
        self.filelist = []

        self._verify_path(dirpath)
        descpath = self._path + "/DESCFILE.INI"
        self.descfile = read_descfile(descpath)

        self.daylist = []
        self.fullData = pd.DataFrame()

        self.completed = False

    def reset(self):
        self.daylist = []
        self.fullData = pd.DataFrame()
        self.completed = False

    def _verify_path(self, path: str) -> None:
        fl = sorted(os.listdir(path))
        if "DESCFILE.INI" not in fl:
            raise FileNotFoundError("No DESCFILE.INI found in this directory.")
        self.filelist = [file for file in fl if verify_datestr(file)]

    def load_df(self, complete_rows=False) -> pd.DataFrame:
        """Reads folder given in DIRPATH and
        transforms data into DF. Saves it in self.fullData
        - complete_rows (bool): if True, completes DFs with Empty Rows by calling
            THIESDayData.complete_empty()
        """
        self.reset()
        for f in self.filelist:
            filepath = f"{self._path}/{f}"
            daydata = THIESDayData(datatype=self._datatype)
            daydata.read_binfile(binpath=filepath, inipath=self.descfile)
            if complete_rows:
                daydata.complete_empty()
            self.daylist.append(daydata)

        self.fullData = sum(self.daylist, start=THIESDayData(self._datatype))

        return self.fullData

    def complete_empty_dates(self):
        if self.completed:
            return
        date_s = add_date_sep(self.filelist[0])
        date_e = add_date_sep(self.filelist[-1])
        d_range = date_range(date_s, date_e)
        for date in d_range:
            if date not in self.fullData.dataDF["Date"].values:
                # Missing day
                new = THIESDayData(self._datatype)
                new.make_empty(self.descfile, date=date)
                self.fullData += new

        self.fullData.sort_by(["Date", "Time"])
        self.completed = True

    def df2csv(self, outpath: str) -> None:
        # if self._datatype == 'av':
        # FORMAT FOR EX FILES ???
        self.fullData.write_csv(outpath)
        print(f"Data written in: {outpath}.csv")

    def read_write(self, outpath: str):
        """Quick version of the read-write process.
        Reads the path given and writes all BIN file data in same CSV
        Does NOT save as DF the data.
        Does NOT complete missing timestamps with empty rows.
        """
        write_header = True
        bcount = 0
        with open(outpath + ".csv", "w") as outfile:
            for i, f in enumerate(self.filelist):
                filepath = f"{self._path}/{f}"
                daydata = THIESDayData(datatype=self._datatype)
                daydata.read_binfile(binpath=filepath, inipath=self.descfile)
                outfile.write(daydata.dataDF.to_csv(header=write_header))
                bcount += daydata.nbytes
                if i == 0:
                    write_header = False
        print(f"Data written in: {outpath}.csv")

    @property
    def dataDF(self):
        return self.fullData.dataDF

    @property
    def shape(self):
        return self.fullData.shape

    @property
    def size(self):
        return len(self.filelist)

    @property
    def parameters(self):
        return self.fullData.parameters

    def __repr__(self) -> str:
        return str(self.fullData)

    def _repr_html_(self):
        return self.fullData



from .types.sftp_client_types import ListfilesArgs, DownloadfilesArgs
from abc import ABC, abstractmethod
from typing import List


class SFTPClientContract(ABC):
    @abstractmethod
    async def list_files(self, args: ListfilesArgs) -> List[str]:
        pass

    @abstractmethod
    async def download_files(self, args: DownloadfilesArgs) -> None:
        pass
